import re
import time

from ..exceptions import AugerException

STATE_POLL_INTERVAL = 10

class AugerBaseApi(object):
    """Auger API base class implements common business object calls."""

    def __init__(
        self, ctx, parent_api,
        object_name=None, object_id=None):
        super(AugerBaseApi, self).__init__()
        self.parent_api = parent_api
        self.parent_id_name = ""
        self.object_id = object_id
        self.object_name = object_name
        self.rest_api = ctx.rest_api
        self._set_api_request_path()
        self.ctx = ctx

    def list(self, params=None):
        params = {} if params is None else params
        if self.parent_api:
            api_request_path = self.parent_api.api_request_path
            params['%s_id' % api_request_path] = self.parent_api.oid
        if self.object_name:
            params['name'] = self.object_name
        return self.rest_api.request_list(
            '%ss' % self.api_request_path, params)

    def properties(self):
        if self.object_id is not None:
            params = {'id': self.object_id}
            if self.parent_id_name:
                params[self.parent_id_name] = self.parent_api.object_id

            return self.rest_api.call('get_%s' % self.api_request_path, params)

        if self.object_name is None:
            raise AugerException(
                'No name or id was specified for %s' % \
                self._get_readable_name())

        alt_name = self.object_name.replace('_', '-')
        for item in iter(self.list()):
            if item['name'] in [self.object_name, alt_name]:
                self.object_id = item.get('id')
                return item

        return None

    def status(self):
        supported_status = ["Cluster", "ClusterTask", "Component",
            "ExperimentSession", "Organization", "Pipeline", "Project",
            "ProjectFile", "SimilarTrialsRequest", "Subscription"]
        if self.object_in_camel_case not in supported_status:
            return self.properties().get(self._get_status_name())
        else:
            return self.rest_api.get_status(
                self.object_in_camel_case, self.oid).\
                get('data').get(self._get_status_name())

    def wait_for_status(self, progress):
        object_readable_name=self._get_readable_name()
        status_value = self.status()
        last_status = ''

        while status_value in progress:
            if status_value != last_status:
                last_status = status_value
                self._log_status(status_value)

            while status_value == last_status:
                time.sleep(STATE_POLL_INTERVAL)
                status_value = self.status()

        if status_value == 'processed_with_error':
            props = self.properties()
            raise AugerException(
                '%s processed with error: %s' % (object_readable_name, props.get('error_message', '')))
        elif status_value == 'error' or status_value == "failure":
            props = self.properties()
            raise AugerException('Auger Cloud return error: %s. Error details: %s'%(props.get('result', ''), props.get('error_message', '')))

        self._log_status(status_value)
        return status_value

    def delete(self):
        self.rest_api.call(
            'delete_%s' % self.api_request_path, {'id': self.oid})

    @property
    def name(self):
        if self.object_name is None:
            properties = self.properties()
            if properties is None:
                raise AugerException(
                    'Can\'t find name for remote %s: %s...' % \
                    (self._get_readable_name(), self.object_id))
            self.object_name = properties.get('name')
        return self.object_name

    @property
    def oid(self):
        return self._ensure_object_id()

    @property
    def is_exists(self):
        return self.properties() != None

    def _get_readable_name(self):
        s = self.api_request_path
        return ' '.join([w.capitalize() for w in s.split('_')])

    def _get_status_name(self):
        return 'status'

    def _log_status(self, status):
        self.ctx.log(
            '%s %s is %s...' % \
            (self._get_readable_name(), self._get_status_name(), status))

    def _call_create(self, params=None, progress=None,has_return_object=True):
        if self.ctx.provider_info:
            provider = list(self.ctx.provider_info.keys())[0]
            params['providers'] = [provider]
            if self.ctx.provider_info[provider].get(self.api_request_path):
                params['provider_info'] = self.ctx.provider_info[provider][self.api_request_path]

        object_properties = self.rest_api.call(
            'create_%s' % self.api_request_path, params)
        if has_return_object:
            if object_properties:
                self.object_id = object_properties.get('id')
                if progress:
                    self.wait_for_status(progress)
            return self.properties()

    def _call_update(self, params=None, progress=None, has_return_object=True):
        if self.ctx.provider_info:
            provider = list(self.ctx.provider_info.keys())[0]
            params['providers'] = [provider]
            if self.ctx.provider_info[provider].get(self.api_request_path):
                params['provider_info'] = self.ctx.provider_info[provider][self.api_request_path]

        #print("_call_update %s: %s"%(self.api_request_path,params))
        object_properties = self.rest_api.call(
            'update_%s' % self.api_request_path, params)

        if has_return_object:
            if object_properties:
                self.object_id = object_properties.get('id')
                if progress:
                    self.wait_for_status(progress)
            return self.properties()

    def _ensure_object_id(self):
        if self.object_id is None:
            properties = self.properties()
            if properties is not None:
                self.object_id = properties.get('id')
            else:
                raise AugerException('Can\'t find remote %s: %s...' % \
                    (self._get_readable_name(), self.object_name))
        return self.object_id

    def _get_uniq_object_name(self, prefix, suffix):
        all_similar_names, count = [], 0
        for item in iter(self.list()):
            if prefix in item.get('name'):
                all_similar_names.append(item.get('name'))
                count += 1

        if count == 0:
            return '%s%s' % (prefix, suffix)

        max_tries = count + 100
        while count < max_tries:
            name = '%s-%s%s' % (prefix, count, suffix)
            if name in all_similar_names:
                count += 1
            else:
                return name

        return '%s-%s%s' % (prefix, shortuuid.uuid(), suffix)

    def _set_api_request_path(self, patch_name=None):
        def to_snake_case(name):
            s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
            return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()
        def to_camel_case(name):
            return ''.join(x.capitalize() for x in name.split('_'))
        def get_api_request_path(name):
            return '_'.join(to_snake_case(name).split('_')[1:-1])

        self.api_request_path = get_api_request_path(
            patch_name if patch_name else type(self).__name__)
        self.object_in_camel_case = to_camel_case(
            self.api_request_path)



import os

from a2ml.api.utils import fsclient
from .deploy import ModelDeploy
from ..cloud.pipeline import AugerPipelineApi
from ..cloud.endpoint import AugerEndpointApi

class ModelUndeploy(object):
    """Undeploy Model locally or from Auger Cloud."""

    def __init__(self, ctx, project):
        super(ModelUndeploy, self).__init__()
        self.project = project
        self.ctx = ctx

    def execute(self, model_id, locally=False):
        if locally:
            is_loaded, model_path, model_name = \
                ModelDeploy(self.ctx, self.project).verify_local_model(model_id)
            self.ctx.log("Undeploy model. Remove local model: %s" % model_name)

            if is_loaded:
                fsclient.remove_file(model_name)

            model_folder = os.path.splitext(model_name)[0]
            fsclient.remove_folder(model_folder)
        else:
            pipeline_api = AugerPipelineApi(self.ctx, None, model_id)
            if pipeline_api.check_endpoint():
                self.ctx.log("Undeploy Review endpoint and all models.")
                endpoint_api = AugerEndpointApi(self.ctx, None, pipeline_api.object_id)
                endpoint_props = endpoint_api.properties()
                for pipeline in endpoint_props.get('endpoint_pipelines', []):
                    try:
                        AugerPipelineApi(self.ctx, None, pipeline.get('pipeline_id')).remove(pipeline.get('pipeline_id'))
                    except Exception as e:
                        if "transition to this status impossible" in str(e):
                            self.ctx.log("Model %s already undeployed."%pipeline.get('pipeline_id'))
                        else:
                            self.ctx.error("Model %s undeploy error: %s"%(pipeline.get('pipeline_id'),e))

                endpoint_api.delete()
            else:
                pipeline_api.remove(model_id)                    




import datetime
import os
import logging
import numpy as np
import json

from a2ml.api.utils import get_uid, get_uid4, fsclient, remove_dups_from_list
from a2ml.api.utils.dataframe import DataFrame


class ModelHelper(object):

    @staticmethod
    def get_root_paths():
        root_dir = os.environ.get('AUGER_ROOT_DIR', '')
        root_tmp_dir = os.environ.get('AUGER_ROOT_TMP_DIR', '')

        if os.environ.get('S3_DATA_PATH'):
            root_dir = os.path.join("s3://", os.environ.get('S3_DATA_PATH'))

            if not os.environ.get('AUGER_LOCAL_TMP_DIR'):
                os.environ["AUGER_LOCAL_TMP_DIR"] = root_tmp_dir

            root_tmp_dir = os.path.join(root_dir, 'temp')
        elif not os.environ.get('AUGER_LOCAL_TMP_DIR'):
            os.environ["AUGER_LOCAL_TMP_DIR"] = root_tmp_dir

        return root_dir, root_tmp_dir

    @staticmethod
    def get_project_path():
        root_dir, root_tmp_dir = ModelHelper.get_root_paths()
        return os.path.join(root_dir, os.environ.get('AUGER_PROJECT_PATH', ''))

    @staticmethod
    def get_models_path(project_path=None):
        if not project_path:
            project_path = ModelHelper.get_project_path()
        return os.path.join(project_path, "models")

    @staticmethod
    def get_model_path(model_id, project_path=None):
        if model_id:
            return os.path.join(ModelHelper.get_models_path(project_path), model_id)

    @staticmethod
    def get_metrics_path(params):
        project_path = params.get('hub_info', {}).get('project_path')
        if not project_path:
            project_path = ModelHelper.get_project_path()

        if params.get('hub_info', {}).get('experiment_id') and params.get('hub_info', {}).get('experiment_session_id'):
            return os.path.join(project_path, "channels", params['hub_info']['experiment_id'],
                "project_runs", params['hub_info']['experiment_session_id'], "metrics")

        return None

    @staticmethod
    def get_metric_path(params, metric_id=None):
        if not metric_id:
            metric_id = params.get('hub_info', {}).get('pipeline_id')

        if not metric_id:
            metric_id = params.get('uid')

        metrics_path = ModelHelper.get_metrics_path(params)
        if metrics_path:
            return os.path.join(metrics_path, metric_id)

        return None

    @staticmethod
    def save_metric(metric_id, project_path, metric_name, metric_data):
        metric_path = ModelHelper.get_metric_path({'hub_info':{'project_path': project_path}}, metric_id)

        fsclient.write_json_file(os.path.join(metric_path,
            "metric_names_feature_importance.json"))

    # @staticmethod
    # def _get_score_byname(scoring):
    #     from sklearn.metrics.scorer import get_scorer
    #     from sklearn.metrics import SCORERS

    #     #TODO: below metrics does not directly map to sklearn:
    #     # Classification : weighted_accuracy, accuracy_table, balanced_accuracy, matthews_correlation,norm_macro_recall
    #     # Regression,  Time Series Forecasting:
    #     #spearman_correlation, normalized_root_mean_squared_error, normalized_mean_absolute_error
    #     scorer = None
    #     if scoring.startswith("AUC"):
    #         scorer = get_scorer("roc_auc")
    #         average = scoring.split("_")[-1]
    #         scorer._kwargs['average'] = average
    #     elif scoring.startswith("log_loss"):
    #         scorer = get_scorer("neg_log_loss")
    #     # elif scoring.startswith("matthews_correlation"):
    #     #     scorer = get_scorer("matthews_corrcoef")
    #     elif scoring.startswith("precision_score"):
    #         scorer = get_scorer("precision")
    #         average = scoring.split("_")[-1]
    #         scorer._kwargs['average'] = average
    #     elif scoring.startswith("average_precision_score"):
    #         scorer = get_scorer("average_precision")
    #         average = scoring.split("_")[-1]
    #         scorer._kwargs['average'] = average
    #     elif scoring.startswith("recall_score"):
    #         scorer = get_scorer("recall")
    #         average = scoring.split("_")[-1]
    #         scorer._kwargs['average'] = average
    #     elif scoring.startswith("norm_macro_recall"):
    #         scorer = get_scorer("recall")
    #         scorer._kwargs['average'] = "macro"
    #     elif scoring.startswith("f1_score"):
    #         scorer = get_scorer("f1")
    #         average = scoring.split("_")[-1]
    #         scorer._kwargs['average'] = average
    #     elif scoring.startswith("precision_score"):
    #         scorer = get_scorer("precision")
    #         average = scoring.split("_")[-1]
    #         scorer._kwargs['average'] = average
    #     elif scoring.startswith("spearman_correlation"):
    #         scorer = get_scorer("r2")
    #     elif scoring.startswith("r2_score"):
    #         scorer = get_scorer("r2")
    #     elif "mean_absolute_error" in scoring:
    #         scorer = get_scorer("neg_mean_absolute_error")
    #     elif "root_mean_squared" in scoring:
    #         scorer = get_scorer("mean_squared_error")
    #     elif "median_absolute_error" in scoring:
    #         scorer = get_scorer("neg_median_absolute_error")

    #     if scorer is None:
    #         scorer = get_scorer(scoring)

    #     return scorer

    @staticmethod
    def calculate_scores(options, y_test, X_test=None, estimator=None, y_pred=None, raise_main_score=True):
        from sklearn.metrics.scorer import get_scorer
        from sklearn.model_selection._validation import _score
        # For calculate_scores
        from .scores.regression import spearman_correlation_score
        from .scores.classification import AUC_weighted_score

        import inspect

        if options.get('fold_group') == 'time_series_standard_model':
            pp_fold_groups_params = options.get('pp_fold_groups_params', {}).get(options['fold_group'], {})
            if pp_fold_groups_params.get('scale_target_min') and pp_fold_groups_params.get('scale_target_max'):
                corr_min = pp_fold_groups_params['scale_target_min']
                corr_max = pp_fold_groups_params['scale_target_max']

                if estimator:
                    y_pred = estimator.predict(X_test)

                y_test = y_test * corr_max + corr_min
                if isinstance(y_pred, list):
                    y_pred = np.array(y_pred)

                y_pred = y_pred * corr_max + corr_min
            else:
                logging.error("calculate_scores: no scaling found for target fold group: %s"%options['fold_group'])

        all_scores = {}
        for scoring in options.get('scoreNames', []):
            try:
                if options.get('task_type') == "timeseries":
                    # from auger_ml.preprocessors.space import ppspace_is_timeseries_model

                    # if ppspace_is_timeseries_model(options.get('algorithm_name')) and \
                    #     scoring != options.get('scoring'):
                    #     continue
                    if scoring != options.get('scoring'):
                        continue

                if scoring == 'r2_score':
                    scoring = 'r2'

                scorer = get_scorer(scoring)
                if options.get('minority_target_class_pos') is not None:
                    argSpec = inspect.getfullargspec(scorer._score_func)
                    if 'pos_label' in argSpec.args:
                        scorer._kwargs['pos_label'] = options.get('minority_target_class_pos')
                        #logging.info("Use minority class to calculate score: %s"%scorer._kwargs)

                if y_pred is not None:
                    all_scores[scoring] = scorer._sign * scorer._score_func(y_test, y_pred, **scorer._kwargs)
                else:
                    all_scores[scoring] = _score(estimator, X_test, y_test, scorer)
                    #all_scores['scoring'] = scorer(estimator, X_test, y_test)


                if np.isnan(all_scores[scoring]):
                    all_scores[scoring] = 0

            except Exception as e:
                #logging.exception("Score failed.")
                if scoring == options.get('scoring', None) and raise_main_score:
                    raise

                logging.error("Score %s for algorithm %s failed to build: %s" % (
                    scoring, options.get('algorithm_name'), str(e)))
                all_scores[scoring] = 0

        return all_scores

    @staticmethod
    def preprocess_target(model_path, data_path=None, records=None, features=None):
        ds = DataFrame.create_dataframe(data_path, records, features)

        return ModelHelper.preprocess_target_ds(model_path, ds)

    @staticmethod
    def preprocess_target_ds(model_path, ds):
        options = fsclient.read_json_file(os.path.join(model_path, "options.json"))
        target_categoricals = fsclient.read_json_file(os.path.join(model_path, "target_categoricals.json"))
        y_true =  None

        if not options.get('targetFeature') or not options.get('targetFeature') in ds.columns:
            return y_true, target_categoricals

        if options.get('timeSeriesFeatures'):
            y_true = np.ravel(ds.df[options.get('targetFeature')].astype(np.float64, copy=False), order='C')
        else:
            if target_categoricals and target_categoricals.get(options.get('targetFeature'), {}).get('categories'):
                ds.convertToCategorical(options.get('targetFeature'), is_target=True,
                    categories=target_categoricals.get(options.get('targetFeature')).get('categories'))

            y_true = np.ravel(ds.df[options.get('targetFeature')], order='C')

        return y_true, target_categoricals

    @staticmethod
    def process_prediction(ds, results, results_proba, proba_classes,
        threshold, minority_target_class, targetFeature, target_categories):

        if results_proba is not None:
            proba_classes_orig = None
            if target_categories:
                proba_classes_orig = ModelHelper.revertCategories(proba_classes, target_categories)

            results = ModelHelper.calculate_proba_target(
                results_proba, proba_classes, proba_classes_orig,
                threshold, minority_target_class)

            if proba_classes_orig is not None:
                proba_classes = proba_classes_orig

        try:
            results = list(results)
        except Exception as e:
            results = [results]

        if target_categories and results_proba is not None:
            results = ModelHelper.revertCategories(results, target_categories)

        # drop target
        if targetFeature in ds.columns:
            ds.drop([targetFeature])

        try:
            results = list(results)
        except Exception as e:
            results = [results]

        ds.df[targetFeature] = results
        if results_proba is not None:
            for idx, name in enumerate(proba_classes):
                ds.df['proba_'+str(name)] = list(results_proba[:, idx])

    @staticmethod
    def save_prediction(ds, prediction_id, support_review_model,
        json_result, count_in_result, prediction_date, model_path, model_id, output=None, gzip_predict_file=False,
        prediction_id_col=None,model_features=None):
        if prediction_id_col is not None:
            ds.df['prediction_id'] = prediction_id_col
        else:
            # Ids for each row of prediction (prediction row's ids)
            prediction_ids = []
            for i in range(0, ds.count()):
                prediction_ids.append(get_uid4())

            ds.df.insert(loc=0, column='prediction_id', value=prediction_ids)

        return ModelHelper.save_prediction_result(ds, prediction_id, support_review_model,
            json_result, count_in_result, prediction_date, model_path, model_id, output, 
            gzip_predict_file=gzip_predict_file, model_features=model_features)

    @staticmethod
    def save_prediction_result(ds, prediction_id, support_review_model,
        json_result, count_in_result, prediction_date, model_path, model_id, output=None, 
        gzip_predict_file=False, model_features=None):
        path_to_predict = ds.options.get('data_path')
        # Id for whole prediction (can contains many rows)
        if not prediction_id:
            prediction_id = get_uid()

        result = {}    
        if path_to_predict and not json_result:
            if output:
                predict_path = output
            else:
                parent_path = os.path.dirname(path_to_predict)
                file_name = os.path.basename(path_to_predict)
                predict_path = os.path.join(parent_path, "predictions",
                    os.path.splitext(file_name)[0] + "_%s_%s_predicted.csv" % (prediction_id, model_id))

            compression = None
            if gzip_predict_file:
                predict_path += ".gz"
                compression = 'gzip'

            ds.saveToCsvFile(predict_path, compression=compression)

            if count_in_result:
                result = {'result_path': predict_path, 'count': ds.count()}
            else:
                result = predict_path
        else:
            if ds.loaded_columns or json_result:
                predicted = ds.df.to_dict('split')
                result = {'data': predicted.get('data', []), 'columns': predicted.get('columns')}
            elif ds.from_pandas:
                result = ds.df
            else:    
                result = ds.df.to_dict('records')

        if support_review_model:
            file_name = str(prediction_date or datetime.date.today()) + \
                '_' + prediction_id + "_results.feather.zstd"
            #Save only model features, they should contain target and prediction_id    
            if model_features:
                ds.select(model_features)

            ds.saveToFeatherFile(os.path.join(
                model_path, "predictions", file_name))

        return result
            
    @staticmethod
    def revertCategories(results, categories):
        return list(map(lambda x: categories[int(x)], results))

    @staticmethod
    def calculate_proba_target(results_proba, proba_classes, proba_classes_orig,
                               threshold, minority_target_class=None):
        results = []

        if type(threshold) == str:
            try:
                threshold = float(threshold)
            except:
                try:
                    threshold = json.loads(threshold)
                except Exception as e:
                    raise Exception("Threshold '%s' should be float or hash with target classes. Error: %s" % (
                        threshold, str(e)))

        if not proba_classes_orig:
            proba_classes_orig = proba_classes

        if type(threshold) != dict:
            if minority_target_class is None:
                minority_target_class = proba_classes_orig[-1]

            threshold = {minority_target_class: threshold}

        mapped_threshold = {}

        for name, value in threshold.items():
            idx_class = None
            for idx, item in enumerate(proba_classes_orig):
                if item == name:
                    idx_class = idx
                    break

            if idx_class is None:
                raise Exception("Unknown target class in threshold: %s, %s" % (
                    name, proba_classes_orig))

            mapped_threshold[idx_class] = value

        for item in results_proba:
            proba_idx = None
            for idx, value in mapped_threshold.items():
                if item[idx] >= value:
                    proba_idx = idx
                    break

            # Find class with value > threshold from the last
            if proba_idx is None and len(mapped_threshold) == 1:
                threshold_value = list(mapped_threshold.values())[0]
                for idx, value in enumerate(item):
                    if item[len(item)-idx-1] >= threshold_value:
                        proba_idx = len(item)-idx-1
                        break

            # Find any class not minority_target_class  from the last
            if proba_idx is None:
                proba_idx = len(item)-1
                for idx, value in enumerate(item):
                    if len(item)-idx-1 not in mapped_threshold:
                        proba_idx = len(item)-idx-1
                        break

            results.append(proba_classes[proba_idx])

        return results

    @staticmethod
    def get_train_features(options):
        selected_cols = []
        for item in options.get('originalFeatureColumns', []):
            selected_cols.append(str(item))

        if options.get('targetFeature') is not None:
            selected_cols.append(options.get('targetFeature'))
        if options.get('timeSeriesFeatures'):
            selected_cols.extend(options.get('timeSeriesFeatures'))

        return remove_dups_from_list(selected_cols)




import os
import logging
import pandas as pd
import datetime as dt
import shortuuid
import json
import re

from azureml.core import Run
from azureml.core import Dataset
from azureml.core import Experiment
from azureml.core.compute import AmlCompute
from azureml.core.compute import ComputeTarget
from azureml.train.automl import AutoMLConfig
from azureml.train.automl.run import AutoMLRun
from azureml.automl.core.featurization import FeaturizationConfig

from .project import AzureProject
from .exceptions import AzureException
from a2ml.api.utils.decorators import error_handler, authenticated
from .credentials import Credentials
from a2ml.api.utils.formatter import print_table
from a2ml.api.azure.dataset import AzureDataset
from a2ml.api.utils import fsclient


class AzureExperiment(object):

    def __init__(self, ctx):
        super(AzureExperiment, self).__init__()
        self.ctx = ctx
        self.credentials = Credentials(self.ctx).load()

    @error_handler
    @authenticated
    def list(self):
        ws = AzureProject(self.ctx)._get_ws()
        experiments = Experiment.list(workspace=ws)
        nexperiments = len(experiments)
        experiments = [e.name for e in experiments]
        for name in experiments:
            self.ctx.log(name)
        self.ctx.log('%s Experiment(s) listed' % str(nexperiments))
        return {'experiments': experiments}

    @staticmethod    
    def _map_metric_a2ml_to_azure(metric):
        if metric == "r2":
            metric = "r2_score"
        elif metric == "precision_weighted":
            metric == "precision_score_weighted"

        return metric

    @staticmethod    
    def _map_metric_azure_to_a2ml(metric):
        if metric == "r2_score":
            metric = "r2"
        elif metric == "precision_score_weighted":
            metric == "precision_weighted"

        return metric

    @error_handler
    @authenticated    
    def start(self):
        model_type = self.ctx.config.get('model_type')
        if not model_type:
            raise AzureException('Please specify model type...')

        if model_type == 'timeseries':
            model_type = 'forecasting'

        primary_metric = self._map_metric_a2ml_to_azure(self.ctx.config.get('experiment/metric'))
        if not primary_metric:
            raise AzureException('Please specify primary metric...')
        #TODO: check if primary_metric is constent with model_type
        target = self.ctx.config.get('target')
        if not target:
            raise AzureException('Please specify target column...')

        dataset_name = self.ctx.config.get('dataset', None)
        if dataset_name is None:
            raise AzureException('Please specify Dataset name...')
        experiment_name = self._fix_experiment_name(
            self.ctx.config.get('experiment/name', dataset_name))

        self.ctx.log("Starting search on %s Dataset..." % dataset_name)
        exclude_columns = self.ctx.config.get_list('exclude', [])
        if target in exclude_columns:
            exclude_columns.remove(target)

        project = AzureProject(self.ctx)    
        ws = project._get_ws()     
        dataset = Dataset.get_by_name(ws, dataset_name)
        if exclude_columns:
            dataset = dataset.drop_columns(exclude_columns)

        compute_target, cluster_name = self._get_compute_target(ws, project)
        automl_settings = {
            "iteration_timeout_minutes" : self.ctx.config.get(
                'experiment/max_eval_time',10),
            "iterations" : self.ctx.config.get(
                'experiment/max_n_trials',10),
            "primary_metric" : primary_metric,
            "verbosity" : logging.INFO,
            "enable_stack_ensemble": self.ctx.config.get(
                'experiment/use_ensemble', False),
            "enable_voting_ensemble": self.ctx.config.get(
                'experiment/use_ensemble', False),
        }

        validation_data = None
        if self.ctx.config.get('experiment/validation_source'):
            if self.ctx.config.get('experiment/validation_dataset'):
                validation_data = Dataset.get_by_name(ws, self.ctx.config.get('experiment/validation_dataset'))
            if not validation_data:
                res = AzureDataset(self.ctx).create(
                    source = self.ctx.config.get('experiment/validation_source'),
                    validation = True
                )

                training_data_columns = AzureDataset(self.ctx, ws)._columns(dataset)
                training_data_columns.remove(target)

                validation_data = Dataset.get_by_name(ws, res['dataset']).keep_columns(training_data_columns)
        else:
            self.ctx.config.remove('experiment/validation_dataset')
            self.ctx.config.write()

            automl_settings["n_cross_validations"] = self.ctx.config.get(
                'experiment/cross_validation_folds', 5)
            if self.ctx.config.get('experiment/validation_size'):
                automl_settings["validation_size"] = self.ctx.config.get('experiment/validation_size')

        if self.ctx.config.get('experiment/max_total_time'):
            automl_settings["experiment_timeout_hours"] = float(self.ctx.config.get('experiment/max_total_time'))/60.0
        if self.ctx.config.get('experiment/max_cores_per_trial'):
            automl_settings["max_cores_per_iteration"] = self.ctx.config.get('experiment/max_cores_per_trial')
        if self.ctx.config.get('experiment/max_concurrent_trials'):
            automl_settings["max_concurrent_iterations"] = self.ctx.config.get('experiment/max_concurrent_trials')
        if self.ctx.config.get('experiment/blocked_models'):
            automl_settings["blocked_models"] = self.ctx.config.get_list('experiment/blocked_models')
        if self.ctx.config.get('experiment/allowed_models'):
            automl_settings["allowed_models"] = self.ctx.config.get_list('experiment/allowed_models')

        # if self.ctx.config.get('exclude'):
        #     fc = FeaturizationConfig()
        #     fc.drop_columns = self.ctx.config.get('exclude').split(",")
        #     automl_settings["featurization"] = fc

        # It should be empty folder
        snapshot_path = os.path.join(os.getcwd(), ".azureml")
        #fsclient.create_folder()
        automl_config = AutoMLConfig(
            task = model_type,
            debug_log = 'automl_errors.log',
            path = snapshot_path,
            compute_target = compute_target,
            training_data = dataset,
            validation_data = validation_data,
            label_column_name = target,
            model_explainability = True, #To get feature importance
            **automl_settings)

        experiment = Experiment(ws, experiment_name)
        run = experiment.submit(automl_config, show_output = False)

        self.ctx.log("Started Experiment %s search..." % experiment_name)
        self.ctx.config.set('experiment/name', experiment_name)
        self.ctx.config.set('cluster/name', cluster_name)
        self.ctx.config.set('experiment/run_id', run.run_id)
        self.ctx.config.write()

        return {'experiment_name': experiment_name, 'run_id': run.run_id}

    @error_handler
    @authenticated    
    def stop(self, run_id = None):
        ws = AzureProject(self.ctx)._get_ws()
        experiment_name = self.ctx.config.get('experiment/name', None)
        if experiment_name is None:
            raise AzureException('Please specify Experiment name...')
        if run_id is None:
            run_id = self.ctx.config.get('experiment/run_id', None)
        if run_id is None:
            raise AzureException(
                'Please provide Run ID (experiment/run_id)...')
        experiment = Experiment(ws, experiment_name)
        run = AutoMLRun(experiment = experiment, run_id = run_id)
        run.cancel()
        return {'stopped': experiment_name}

    @error_handler
    @authenticated    
    def leaderboard(self, run_id = None):
        ws = AzureProject(self.ctx)._get_ws()
        experiment_name = self.ctx.config.get('experiment/name', None)
        if experiment_name is None:
            raise AzureException('Please specify Experiment name...')
        if run_id is None:
            run_id = self.ctx.config.get('experiment/run_id', None)
        if run_id is None:
            raise AzureException(
                'Please provide Run ID (experiment/run_id) to evaluate')
        experiment = Experiment(ws, experiment_name)
        run = AutoMLRun(experiment = experiment, run_id = run_id)
        leaderboard, trials_count = self._get_leaderboard(run)
        leaderboard = leaderboard.to_dict('records')
        self.ctx.log('Leaderboard for Run %s' % run_id)

        headers = []
        if leaderboard:
            headers = list(leaderboard[0].keys())[:3]
        print_table(self.ctx.log, leaderboard, headers)
        provider_status = run.get_status()
        status = self._map_provider_status(provider_status)

        result = {
            'run_id': run_id,
            'leaderboard': leaderboard,
            'trials_count': trials_count,
            'status': status,
            'provider_status': provider_status,
        }

        if status == 'error':
            result['error'] = run.properties.get('errors')
            result['error_details'] = run.get_details().get('error', {}).get('error', {}).get('message')
            self.ctx.log('Status: %s, Error: %s, Details: %s' % (
                status, result['error'], result['error_details']
            ))
            self.ctx.log_debug(run.get_details().get('error'))
        else:    
            self.ctx.log('Status: %s' % status)

        return result
            
    def _map_provider_status(self, provider_status):
        # * NotStarted - This is a temporary state client-side Run objects are in before cloud submission.
        # * Starting - The Run has started being processed in the cloud. The caller has a run ID at this point.
        # * Provisioning - Returned when on-demand compute is being created for a given job submission.
        # * Preparing - The run environment is being prepared:
        #     * docker image build
        #     * conda environment setup
        # * Queued - The job is queued in the compute target. For example, in BatchAI the job is in queued state
        #      while waiting for all the requested nodes to be ready.
        # * Running - The job started to run in the compute target.
        # * Finalizing - User code has completed and the run is in post-processing stages.
        # * CancelRequested - Cancellation has been requested for the job.
        # * Completed - The run completed successfully. This includes both the user code and run
        #     post-processing stages.
        # * Failed - The run failed. Usually the Error property on a run will provide details as to why.
        # * Canceled - Follows a cancellation request and indicates that the run is now successfully cancelled.
        # * NotResponding - For runs that have Heartbeats enabled, no heartbeat has been recently sent.


        if provider_status == 'NotStarted' or provider_status == 'Starting' or \
           provider_status == 'Provisioning' or provider_status == 'Preparing' or \
           provider_status == 'Queued':
            return "preprocess"

        if provider_status == 'Running' or provider_status == 'Finalizing':
            return "started"

        if provider_status == 'Completed':
            return "completed"

        if provider_status == 'Failed':
            return "error"

        if provider_status == 'CancelRequested' or provider_status == 'Canceled':
            return "interrupted"
                
    @error_handler 
    @authenticated       
    def get_experiment_settings(self):
        return 
                
    @error_handler 
    @authenticated       
    def history(self):
        ws = AzureProject(self.ctx)._get_ws()
        experiment_name = self.ctx.config.get('experiment/name', None)
        if experiment_name is None:
            raise AzureException('Please specify Experiment name...')
        experiment = Experiment(ws, experiment_name)
        runs = Run.list(experiment)
        result = []
        for run in runs:
            details = run.get_details()
            st = dt.datetime.strptime(
                details['startTimeUtc'], '%Y-%m-%dT%H:%M:%S.%fZ')
            et = dt.datetime.strptime(
                details['endTimeUtc'], '%Y-%m-%dT%H:%M:%S.%fZ')
            duratin = str(et-st)
            result.append({
                'id': run.id,
                'start time': details['startTimeUtc'],
                'duratin': duratin,
                'status': details['status']
            })
        print_table(self.ctx.log, result)
        return {'history': result}

    def _fix_experiment_name(self, name):
        # Experiment name must be between 1 and 255 characters long. 
        # Its first character has to be alphanumeric, and the rest may contain hyphens and underscores. 
        # No whitespace is allowed.

        name = re.sub(r'\W+', '-', name)
        name = name[:255]

        return name

    def _get_compute_target(self, ws, project):
        local_cluster = project.get_cluster_config(name=None, local_config=True, ws=ws)
        project.update_cluster_config(name=None, params=local_cluster, ws=ws,
            allow_create=not self.ctx.is_runs_on_server())

        return ws.compute_targets[local_cluster['name']], local_cluster['name']

    def _get_leaderboard(self, experiment_run):
        primary_metric = self._map_metric_azure_to_a2ml(experiment_run.properties['primary_metric'])
        task_type = ""
        if experiment_run.properties.get("AMLSettingsJsonString"):
            settings = json.loads(experiment_run.properties.get("AMLSettingsJsonString"))
            task_type = settings.get('task_type')

        children = list(experiment_run.get_children(recursive=True))
        leaderboard = pd.DataFrame(index=['model id', 'algorithm', 'score', 'fit_time', 'algorithm_name', 'algorithm_params', 'preprocessor', 'primary_metric', "all_scores", 'task_type'])
        goal_minimize = False
        for run in children:
            run_props = run.properties
            if('run_algorithm' in run_props and 'score' in run_props):
                if run_props['run_preprocessor']:
                    run_algorithm = '%s,%s' % (run_props['run_preprocessor'],
                        run_props['run_algorithm'])
                else:
                    run_algorithm = run_props['run_algorithm']

                algorithm_params = {}    
                if run_props.get('pipeline_spec'):
                    pipeline_spec = json.loads(run_props.get('pipeline_spec'))
                    for item in pipeline_spec.get('objects', []):
                        if item.get('spec_class') and item.get('spec_class') != "preproc" and \
                            not "Ensemble" in item.get('class_name', ""):
                            algorithm_params = item.get('param_kwargs')
                            break

                all_scores_azure = run.get_metrics()
                all_scores = {}
                scores_to_remove = ['confusion_matrix', 'accuracy_table', 'predicted_true', 'residuals']
                for key, value in all_scores_azure.items():
                    if key in scores_to_remove:
                        continue

                    all_scores[self._map_metric_azure_to_a2ml(key)] = value    

                leaderboard[run.id] = [
                    run.id,
                    run_algorithm,
                    float(run_props['score']),
                    run_props['fit_time'],
                    run_props['run_algorithm'],
                    algorithm_params,
                    run_props['run_preprocessor'],
                    primary_metric,
                    all_scores,
                    task_type
                ]
                if('goal' in run_props):
                    goal_minimize = run_props['goal'].split('_')[-1] == 'min'
                    
        leaderboard = leaderboard.T.sort_values(
            'score', ascending = goal_minimize)
        trials_count = len(leaderboard)
        leaderboard = leaderboard.head(10)
        leaderboard.rename(columns={'score':primary_metric}, inplace=True)

        return leaderboard, trials_count



import os
import pandas as pd
import datetime
import math
import copy
import logging

from a2ml.api.utils import get_uid, convert_to_date, merge_dicts, fsclient
from a2ml.api.utils.dataframe import DataFrame

from .model_helper import ModelHelper
from .probabilistic_counter import ProbabilisticCounter


class ModelReview(object):
    def __init__(self, params):
        self.model_path = params.get('model_path')
        if not self.model_path:
            self.model_path = ModelHelper.get_model_path(params['hub_info']['pipeline_id'],
                params['hub_info'].get('project_path'))

        self.options = fsclient.read_json_file(os.path.join(self.model_path, "options.json"))
        if params.get('hub_info'):
            self.options['hub_info'] = params['hub_info']

        self.target_feature = self.options.get('targetFeature')

    # def get_actuals_statistic(self, date_from=None, date_to=None):
    #     count_actuals = self.count_actuals_by_prediction_id()
    #     performance_daily = self.score_model_performance_daily(date_from, date_to)
    #     distribution_chart_stats = self.distribution_chart_stats(date_from, date_to)

    #     return {
    #         'count_actuals': count_actuals,
    #         'performance_daily': performance_daily,
    #         'distribution_chart_stats': distribution_chart_stats
    #     }

    def get_actuals_score(self):
        #TODO: calc score for the all actuals (use some size or count limit)
        return {}

    def _process_actuals(self, ds_actuals,
            prediction_group_id=None, primary_prediction_group_id=None, primary_model_path=None,
            actual_date=None, actuals_id = None, calc_score=False, raise_not_found=False):

        ds_actuals.df.rename(columns={"actual": 'a2ml_actual'}, inplace=True)

        actuals_count = ds_actuals.count()

        primary_ds = None
        if primary_prediction_group_id:
            files = ModelReview._get_prediction_files(primary_model_path, primary_prediction_group_id)
            for (_, df) in DataFrame.load_from_files(files, features=['prediction_id']):
                primary_ds = df
                # should be only one file
                break

        prediction_files = ModelReview._get_prediction_files(self.model_path, prediction_group_id)

        combined_df = None
        for (file, df_prediction_results) in DataFrame.load_from_files(prediction_files):
            if primary_ds is not None:
                ds_actuals.df['prediction_id'] = ModelReview._map_primary_prediction_id_to_candidate(
                    ds_actuals.df['prediction_id'],
                    primary_ds.df['prediction_id'],
                    df_prediction_results.df['prediction_id']
                )

            intersect_df = ds_actuals.df.merge(df_prediction_results.df, on="prediction_id", how="inner")
            if combined_df is None:
                combined_df = intersect_df
            else:
                combined_df = combined_df.append(intersect_df)

            match_count = len(combined_df)
            #TODO: why we check for primary_ds here?
            if actuals_count == match_count or primary_ds is not None:
                break

        #TODO: why we check for primary_ds here?
        if raise_not_found and actuals_count != match_count and primary_ds is None:
            df_diff = ds_actuals.df[['prediction_id']].merge(combined_df[['prediction_id']], 
                on="prediction_id", how="left", indicator=True)
            diff_ids = df_diff[df_diff._merge=='left_only']['prediction_id'].head(100).values.tolist()
            raise Exception("Actual Prediction ID(s) not found in model predictions: %s"%diff_ids)

        result = True
        if calc_score:
            result = self._do_score_actual(combined_df)

        ds_actuals.df = combined_df
        return result

    def _do_score_actual(self, df_data):
        ds_true = DataFrame({})
        ds_true.df = df_data[['a2ml_actual']].rename(columns={'a2ml_actual':self.target_feature})

        ds_predict = DataFrame({})
        ds_predict.df = df_data[[self.target_feature]] # copy to prevent source data modification

        y_pred, _ = ModelHelper.preprocess_target_ds(self.model_path, ds_predict)
        y_true, _ = ModelHelper.preprocess_target_ds(self.model_path, ds_true)

        return ModelHelper.calculate_scores(self.options, y_test=y_true, y_pred=y_pred, raise_main_score=False)

    # prediction_group_id - prediction group for these actuals
    # primary_prediction_group_id - means that prediction_group_id is produced by a candidate model
    # and prediction rows id should be matched with actuals using primary_prediction_group
    def add_actuals(self, actuals_path = None, actual_records=None,
            prediction_group_id=None, primary_prediction_group_id=None, primary_model_path=None,
            actual_date=None, actuals_id = None, calc_score=True, return_count=False):

        features = None
        if actuals_path or (actual_records and type(actual_records[0]) == list):
            features = ['prediction_id', 'actual']

        ds_actuals = DataFrame.create_dataframe(actuals_path, actual_records,
            features=features)
        if features is None:
            ds_actuals.select(['prediction_id', 'actual'])

        actuals_count = ds_actuals.count()
        result = self._process_actuals(ds_actuals, prediction_group_id, primary_prediction_group_id, primary_model_path,
            actual_date, actuals_id, calc_score, raise_not_found=True)

        #print(ds_actuals)
        #ds_actuals.drop(self.target_feature)
        ds_actuals.df = ds_actuals.df.rename(columns={self.target_feature: 'a2ml_predicted'})
        ds_actuals.df = ds_actuals.df.rename(columns={'a2ml_actual':self.target_feature})

        if not actuals_id:
            actuals_id = get_uid()

        file_name = str(actual_date or datetime.date.today()) + '_' + actuals_id + "_actuals.feather.zstd"
        ds_actuals.saveToFeatherFile(os.path.join(self.model_path, "predictions", file_name))

        if return_count:
            return {'score': result, 'count': actuals_count}
        else:
            return result

    def delete_actuals(self, with_predictions=False, begin_date=None, end_date=None):
        if with_predictions and not begin_date and not end_date:
            self.clear_model_results_and_actuals()
        else:
            path_suffix = "_*_actuals.feather.zstd"
            if with_predictions:
                path_suffix = "_*_*.feather.zstd"

            for (curr_date, files) in ModelReview._prediction_files_by_day(self.model_path, begin_date, end_date, path_suffix):
                for file in files:
                    path = file if type(file) == str else file['path']
                    fsclient.remove_file(path)

    def build_review_data(self, data_path=None, output=None):
        if not data_path:
            data_path = self.options['data_path']

        train_features = ModelHelper.get_train_features(self.options)
        ds_train = DataFrame.create_dataframe(data_path, features=train_features)

        all_files = fsclient.list_folder(os.path.join(self.model_path, "predictions/*_actuals.feather.zstd"),
            wild=True, remove_folder_name=False, meta_info=True)
        all_files.sort(
            key=lambda f: f['last_modified'],
            reverse=True
        )

        for (file, ds_actuals) in DataFrame.load_from_files(all_files):
            if not ds_actuals.df.empty:
                ds_actuals.drop(['prediction_id', 'a2ml_predicted'])

                ds_train.df = pd.concat([ds_train.df, ds_actuals.df[ds_train.columns]], ignore_index=True)
                ds_train.drop_duplicates()

        if not output:
            output = os.path.splitext(data_path)[0] + "_review_%s.parquet"%(get_uid())

        ds_train.saveToFile(output)
        return output

    # def count_actuals_by_prediction_id(self, date_from=None, date_to=None):
    #     res = {}
    #     features = ['prediction_group_id', 'prediction_id', self.target_feature]
    #     counter = ProbabilisticCounter()

    #     all_files = fsclient.list_folder(
    #         os.path.join(self.model_path, "predictions/*_actuals.feather.zstd"),
    #         wild=True,
    #         remove_folder_name=False,
    #         meta_info=False
    #     )

    #     #print(all_files)

    #     for (curr_date, files) in ModelReview._prediction_files_by_day(
    #         self.model_path, date_from, date_to, "_*_actuals.feather.zstd"):

    #         for (file, df) in DataFrame.load_from_files(files, features):
    #             ModelReview._remove_duplicates_by(df, 'prediction_id', counter)

    #             agg = df.df.groupby(['prediction_group_id', 'prediction_id']).count()
    #             agg[self.target_feature] = 1 # exclude duplication prediction_id's inside groups
    #             agg = agg.groupby('prediction_group_id').count()

    #             for prediction_group_id, row, in agg.iterrows():
    #                 count = row[0]

    #                 if prediction_group_id not in res:
    #                     res[prediction_group_id] = count
    #                 else:
    #                     res[prediction_group_id] = res[prediction_group_id] + count

    #     return res

    def statistic_daily(self, date_from, date_to):
        features = ['prediction_id', self.target_feature]
        res = {}

        for (curr_date, files) in ModelReview._prediction_files_by_day(
                self.model_path, date_from, date_to, "_*_actuals.feather.zstd"):
            df_actuals = DataFrame({})
            for (file, df) in DataFrame.load_from_files(files, features):
                df_actuals.df = pd.concat([df_actuals.df, df.df])

            res[str(curr_date)] = {}
            res[str(curr_date)]['actuals_count'] = df_actuals.count()

            if df_actuals.count()>0:
                df_actuals.drop_duplicates(['prediction_id'])
            res[str(curr_date)]['actuals_count_unique'] = df_actuals.count()

        for (curr_date, files) in ModelReview._prediction_files_by_day(
                self.model_path, date_from, date_to, "_*_results.feather.zstd"):
            df_results = DataFrame({})
            for (file, df) in DataFrame.load_from_files(files, features):
                df_results.df = pd.concat([df_results.df, df.df])

            if not str(curr_date) in res:
                res[str(curr_date)] = {}

            res[str(curr_date)]['predictions_count'] = df_results.count()
            if df_results.count()>0:
                df_results.drop_duplicates(['prediction_id'])
            res[str(curr_date)]['predictions_count_unique'] = df_results.count()

        return res

    # date_from..date_to inclusive
    def score_model_performance_daily(self, date_from, date_to):
        features = ['prediction_id', self.target_feature, 'a2ml_predicted']
        res = {}

        for (curr_date, files) in ModelReview._prediction_files_by_day(
                self.model_path, date_from, date_to, "_*_actuals.feather.zstd"):
            df_actuals = DataFrame({})
            for (file, df) in DataFrame.load_from_files(files, features):
                df_actuals.df = pd.concat([df_actuals.df, df.df])

            if df_actuals.count() > 0:
                df_actuals.drop_duplicates(['prediction_id'])

                df_actuals.df.rename(columns={self.target_feature: 'a2ml_actual'}, inplace=True)
                df_actuals.df.rename(columns={'a2ml_predicted': self.target_feature}, inplace=True)

                scores = self._do_score_actual(df_actuals.df)
                res[str(curr_date)] = scores[self.options.get('score_name')]

        return res

    def distribution_chart_stats(self, date_from, date_to):
        features = [self.target_feature]
        categoricalFeatures = self.options.get('categoricalFeatures', [])
        mapper = {}
        mapper[self.target_feature] = 'actual_%s'%self.target_feature

        actuals_stats = self._distribution_stats(
            date_from, date_to, "_*_actuals.feather.zstd", features, categoricalFeatures, mapper
        )

        features += self.options.get('originalFeatureColumns', [])
        mapper[self.target_feature] = 'predicted_%s'%self.target_feature
        features_stats = self._distribution_stats(
            date_from, date_to, "_*_results.feather.zstd", features, categoricalFeatures, mapper
        )

        return merge_dicts(features_stats, actuals_stats)

    def set_support_review_model_flag(self, flag_value):
        path = os.path.join(self.model_path, "options.json")

        self.options['support_review_model'] = bool(flag_value)

        fsclient.remove_file(path)
        fsclient.write_json_file(path, self.options, atomic=True)

    def remove_model(self):
        fsclient.remove_folder(self.model_path)
        return True

    def clear_model_results_and_actuals(self):
        fsclient.remove_folder(os.path.join(self.model_path, "predictions"))
        return True

    def _distribution_stats(self, date_from, date_to, path_suffix, features,
        categoricalFeatures=[], feature_mapper={}):
        res = {}
        feature_importances = self._get_feature_importances()
        counter = ProbabilisticCounter()
        second_pass_counter = ProbabilisticCounter()

        for (curr_date, files) in ModelReview._prediction_files_by_day(self.model_path, date_from, date_to, path_suffix):
            stats = {}

            for feature in features:
                stats[feature] = {
                    'count': 0,
                    'sum': 0,
                    'sq_sum': 0,
                    'dist': None,
                    'imp': feature_importances.get(feature, 0)
                }

            df_list = []
            for (file, df) in DataFrame.load_from_files(files, features + ['prediction_id']):
                ModelReview._remove_duplicates_by(df, 'prediction_id', counter)

                df_list.append(df)

            # First pass: calc sum and count in each column for average
            for df in df_list:
                for feature in features:
                    stats[feature]['count'] += df.df[feature].count()

                    if df.df[feature].dtype.name in ['category', 'string', 'object'] or \
                        feature in categoricalFeatures:
                        stats[feature]['dist'] = merge_dicts(
                            stats[feature]['dist'] or {},
                            dict(df.df[feature].value_counts()),
                            lambda v, ov: v + ov
                        )
                    else:
                        stats[feature]['sum'] += df.df[feature].sum()

            # Calc average
            for feature in features:
                if stats[feature]['count'] > 0 and stats[feature]['dist'] == None:
                    stats[feature]['average'] = stats[feature]['sum'] / stats[feature]['count']

            # Second pass: sum of squares of value and average for std dev
            for df in df_list:
                ModelReview._remove_duplicates_by(df, 'prediction_id', second_pass_counter)

                for feature in features:
                    if 'average' in stats[feature]:
                        avg = stats[feature]['average']
                        stats[feature]['sq_sum'] += ((df.df[feature] - avg)**2).sum()

            # Calc std dev
            if len(files) > 0:
                res[str(curr_date)] = ModelReview._calc_stddev_for_features(stats, features, feature_mapper)

        return res

    def _get_feature_importances(self):
        cache_path = ModelHelper.get_metric_path(self.options)

        importance_data = None
        if cache_path:
            importance_data = fsclient.read_json_file(os.path.join(cache_path, "metrics.json")).get('feature_importance_data')
            if not importance_data:
                importance_data = fsclient.read_json_file(os.path.join(cache_path, "metric_names_feature_importance.json")).get('feature_importance_data')

        if importance_data:
            return dict(zip(importance_data['features'], importance_data['scores']))
        else:
            logging.warn("No feature importance in cache: for model %s" % (cache_path))
            return {}

    @staticmethod
    def _calc_stddev_for_features(stats, features, feature_mapper):
        res = {}

        for feature in features:
            count = stats[feature]['count']

            if count > 0:
                feature_key = feature_mapper.get(feature, feature)
                if 'average' in stats[feature]:
                    res[feature_key] = {
                        "avg": stats[feature]['average'],
                        "std_dev": math.sqrt(stats[feature]['sq_sum'] / (count - 1)) if count > 1 else 0
                    }
                else:
                    res[feature_key] = { 'dist': stats[feature]['dist'] }

                res[feature_key]['imp'] = round(stats[feature]['imp'], 6)

        return res

    @staticmethod
    def _prediction_files_by_day(model_path, date_from, date_to, path_suffix):
        if (date_from and not date_to) or (not date_from and date_to):
            # TODO: list all files by suffix, sort them by prefix date and return range of files
            raise Exception("Arguments error: please provide both start and end dates or do not pass any.")

        if date_from:
            date_from = convert_to_date(date_from)
            date_to = convert_to_date(date_to)

            curr_date = date_from

            while curr_date <= date_to:
                path = os.path.join(model_path, "predictions/" + str(curr_date) + path_suffix)
                files = fsclient.list_folder(path, wild=True, remove_folder_name=False, meta_info=False)
                yield (curr_date, files)
                curr_date += datetime.timedelta(days=1)
        else:
            path = os.path.join(model_path, "predictions/" + "*" + path_suffix)
            files = fsclient.list_folder(path, wild=True, remove_folder_name=False, meta_info=False)
            yield ("today", files)

    @staticmethod
    def _remove_duplicates_by(df, column_name, counter):
        dup_flag_column_name = '__duplicate'
        df.df[dup_flag_column_name] = df.df[column_name]
        df.df[dup_flag_column_name] = df.df[dup_flag_column_name].apply(lambda uid: not counter.add(uid))

        df.df = df.df[df.df[dup_flag_column_name] == False]
        df.drop([dup_flag_column_name])

        return df

    @staticmethod
    def _map_primary_prediction_id_to_candidate(prediction_id, primary_prediction_id, candidate_prediction_id):
        primary_prediction_id = primary_prediction_id.rename('primary_prediction_id')

        if len(primary_prediction_id) != len(candidate_prediction_id):
            raise Exception("Primary prediction's rows count doesn't match candidate's one")

        mapper = {}
        for (i, row) in pd.concat([primary_prediction_id, candidate_prediction_id], axis=1).iterrows():
            mapper[row['primary_prediction_id']] = row['prediction_id']


        return prediction_id.map(mapper)

    @staticmethod
    def _get_prediction_files(model_path, prediction_group_id=None):
        predictions_path = os.path.join(model_path, "predictions/*_results.feather.zstd")

        if prediction_group_id:
            predictions_path = os.path.join(model_path, f"predictions/*_{prediction_group_id}_results.feather.zstd")

        files = fsclient.list_folder(predictions_path, wild=True, remove_folder_name=False, meta_info=True)
        files.sort(key=lambda f: f['last_modified'], reverse=True)

        if len(files) == 0:
            raise Exception('there is no prediction results for this model in ' + predictions_path)

        return files



from .impl.cloud.rest_api import RestApi
from .impl.decorators import with_project
from a2ml.api.utils.decorators import error_handler, authenticated
from .impl.model import Model
from .credentials import Credentials

class AugerModel(object):

    def __init__(self, ctx):
        self.ctx = ctx
        self.credentials = Credentials(ctx).load()
        self.ctx.rest_api = RestApi(
            self.credentials.api_url, self.credentials.token)

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def deploy(self, project, model_id, locally, review):
        model_id = Model(self.ctx, project).deploy(model_id, locally, review)
        return {'model_id': model_id}

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def predict(self, project, filename, model_id, threshold, locally, data, columns, predicted_at, output):
        predicted = Model(self.ctx, project).predict(
            filename, model_id, threshold, locally, data, columns, predicted_at, output)

        if filename:
            self.ctx.log('Predictions stored in %s' % predicted)
        
        return {'predicted': predicted}

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def actuals(self, project, model_id, filename=None, actual_records=None, actuals_at=None, locally=False):
        return Model(self.ctx, project).actuals(model_id, filename, actual_records, actuals_at, locally)

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def delete_actuals(self, project, model_id, with_predictions=False, begin_date=None, end_date=None, locally=False):
        return Model(self.ctx, project).delete_actuals(model_id, with_predictions, begin_date, end_date, locally)

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def review_alert(self, project, model_id, parameters):
        return Model(self.ctx, project).review_alert(model_id, parameters)

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def build_review_data(self, project, model_id, locally, output):
        return Model(self.ctx, project).build_review_data(model_id, locally, output)

    @error_handler
    @authenticated
    @with_project(autocreate=False)
    def review(self, project, model_id):
        return Model(self.ctx, project).review(model_id)

    @error_handler
    @authenticated
    @with_project(autocreate=False)    
    def undeploy(self, project, model_id, locally):
        Model(self.ctx, project).undeploy(model_id, locally)
        return {'model_id': model_id}




from .base import AugerBaseApi
from ..exceptions import AugerException
from .review_alert_item import AugerReviewAlertItemApi
from .review_alert import AugerReviewAlertApi


class AugerEndpointApi(AugerBaseApi):
    """Auger Endpoint API."""

    def __init__(self, ctx, endpoint_api, endpoint_id=None):
        super(AugerEndpointApi, self).__init__(
            ctx, endpoint_api, None, endpoint_id)

    def create(self, pipeline_id):
        return self._call_create({'pipeline_id': pipeline_id},[])



from .base import AugerBaseApi
from .cluster_task import AugerClusterTaskApi


class AugerActualApi(AugerBaseApi):
    """Auger Trial API."""

    def __init__(self, ctx, pipeline_api, prediction_id=None, use_endpoint=False):
        super(AugerActualApi, self).__init__(
            ctx, pipeline_api, prediction_id)
        self.use_endpoint = use_endpoint
        if self.use_endpoint:
            self.parent_id_name = "endpoint_id"
            self._set_api_request_path("AugerEndpointActualApi")

    def create(self, records, actuals_at, actuals_path=None):
        params = {}
        if self.use_endpoint:
            params['endpoint_id'] = self.parent_api.object_id
            #params['prediction_group_id'] = 'dffe01f4-39bd-4f7b-832f-61051f8c7e1b'
        else:
            params['pipeline_id'] = self.parent_api.object_id

        if records:
            params['actuals'] = records

        if actuals_path:
            params['actuals_path'] = actuals_path

        if actuals_at:
            params['actuals_at'] = str(actuals_at)

        return self._call_create(
            params=params, 
            has_return_object=False)

    def delete(self, with_predictions, begin_date, end_date):
        params = {
            'with_predictions': with_predictions,
            'from': begin_date,
            'to': end_date
        }
        if self.use_endpoint:
            params['endpoint_id'] = self.parent_api.object_id
        else:
            params['pipeline_id'] = self.parent_api.object_id

        res = self.rest_api.call( 'delete_%ss' % self.api_request_path, params)
        if isinstance(res, list):
            res = res[0]
        cluster_task = AugerClusterTaskApi(self.ctx, cluster_task_id=res['id'])
        cluster_task.wait_for_status(['pending', 'received', 'started', 'retry'])

        return True



import os

class Config:
    def __init__(self):
        self.debug = os.environ.get('DEBUG', 'true').lower() == 'true'
        self.broker_url = os.environ.get('BROKER_URL', 'amqp://localhost/vhost')
        self.task_result_queue = os.environ.get('TASK_RESULT_QUEUE', 'task_result')



import os
import json
import csv
import sys
import inspect
from google.cloud.automl_v1 import AutoMlClient, PredictionServiceClient
#from google.cloud.automl_v1 import enums
import google.auth
from a2ml.api.utils.config import Config
from google.auth.transport.requests import AuthorizedSession

class GoogleA2ML(object):
    """Google A2ML implementation."""

    def __init__(self,ctx):
        super(GoogleA2ML, self).__init__()
        self.ctx = ctx
        self.client = AutoMlClient()
        self.name = ctx.config.get('name',None)
        self.project_id = ctx.config.get('project',None)
        self.compute_region = ctx.config.get('cluster/region','us-central1')
        self.metric = ctx.config.get('experiment/metric',"MINIMIZE_MAE")
        self.project_location = self.client.location_path(self.project_id,self.compute_region)
        self.dataset_id = ctx.config.get('dataset_id',None)
        self.dataset_name = ctx.config.get('dataset_name',None)
        self.source = ctx.config.get('source', None)
        self.dataset_name = self.client.dataset_path(self.project_id, self.compute_region, self.dataset_id)
        self.target = ctx.config.get('target',None)
        self.exclude = ctx.config.get('exclude',None)
        self.max_total_time = ctx.config.get('max_total_time',60)
        self.operation_name = ctx.config.get('operation_name',None)
        self.model_name = ctx.config.get('model_name',None)
        self.gsbucket = ctx.config.get('gsbucket','gs://a2ml')

    def import_data(self):
        self.ctx.log('Google Import Data')
        self.ctx.log("Creating dataset for project location: {}".format(self.project_location))
        if self.source is None:
            self.ctx.log("Please specify a source (URL or local file)")
            return

        create_dataset_response = self.client.create_dataset(
            self.project_location,
            {'display_name': self.name,
            'tables_dataset_metadata': {}})
        self.dataset_id = create_dataset_response.name.split('/')[-1]
        self.ctx.config.set('dataset_id', self.dataset_id)
        self.ctx.config.write()

        print ("Dataset ID: {}".format(self.dataset_id))
        self.dataset_name = self.client.dataset_path(self.project_id, self.compute_region, self.dataset_id)
        self.ctx.log('Dataset name: {}'.format(self.dataset_name))

        if os.path.isfile(self.source):
            print("Copying {} to {}".format(self.source,self.gsbucket))
            cmd = "gsutil cp " + self.source + " "+ self.gsbucket
            result = os.system(cmd)
            print("Result of local copy to Google Storage bucket: {}".format(result))
            self.source = self.gsbucket + "/" + os.path.basename(self.source)

        if self.source.startswith('bq'):
            input_config = {"bigquery_source": {"input_uri": self.source}}
        else:
            # Get the multiple Google Cloud Storage URIs.
            input_uris = self.source.split(",")
            input_config = {"gcs_source": {"input_uris": input_uris}}

        self.ctx.log("Processing import for dataset ID: {}".format(self.dataset_id))

        operation= self.client.import_data(self.dataset_name, input_config)
        # synchronous check of operation status.
        import_data_response = operation.result()
        self.ctx.log("Imported data: {}".format(import_data_response))

        return self

    def train(self,synchronous=False):
        self.ctx.log('Google Train')
        self.ctx.log("Training model: {}".format(self.name))

        self.ctx.log("Listing tables from: {}".format(self.dataset_name))
        list_table_specs_response = self.client.list_table_specs(self.dataset_name)
        self.ctx.log("List table specs response: {}".format(list_table_specs_response))
        table_specs = [s for s in list_table_specs_response]
        table_spec_name = table_specs[0].name
        self.ctx.log("Table spec name: {}".format(table_spec_name))

        list_column_specs_response = self.client.list_column_specs(table_spec_name)
        self.column_specs = {s.display_name: s for s in list_column_specs_response}

        label_column_name = self.target
        label_column_spec = self.column_specs[label_column_name]
        label_column_id = label_column_spec.name.rsplit('/', 1)[-1]
        update_dataset_dict = {
            'name': self.dataset_name,
            'tables_dataset_metadata': {'target_column_spec_id': label_column_id}}
        update_dataset_response = self.client.update_dataset(update_dataset_dict)
        self.ctx.log("Updated dataset response: {}".format(update_dataset_response))
        self.feat_list = list(self.column_specs.keys())
        self.feat_list.remove(self.target)
        excluded = self.exclude.split(',')
        for exclude in excluded:
            self.ctx.log("Removing: {}".format(exclude))
            try:
                self.feat_list.remove(exclude)
            except Exception as inst:
                self.ctx.log("Can't find: {}: {}".format(exclude,inst))

        model_dict = {
        'display_name': self.name,
        'dataset_id': self.dataset_name.rsplit('/',1)[-1],
        'tables_model_metadata': {
            'target_column_spec': self.column_specs[self.target],
            'input_feature_column_specs': [
                self.column_specs[x] for x in self.feat_list],
            'train_budget_milli_node_hours': int(self.max_total_time/60*1000), # budget is in minutes, google wants "millihours", seriously?
            'optimization_objective': self.metric}}
        response = self.client.create_model(self.project_location,model_dict)

        self.operation_name = response.operation.name
        self.ctx.log("Training operation name: {}".format(self.operation_name))
        self.ctx.config.set('operation_name', self.operation_name)

        if synchronous:
            model_response=response.result()
            metadata = model_response.metadata()
            self.ctx.log("Training completed: {}".format(metadata))
            model_name = model_response.name()
            self.ctx.log("Model name: {}".format(model_name))

        self.ctx.config.write()

    def evaluate(self):
        credentials, project = google.auth.default(scopes=['https://www.googleapis.com/auth/cloud-platform'])
        authed_session = AuthorizedSession(credentials)
        basename="https://automl.googleapis.com/v1beta1/"
        cmd = basename + self.operation_name
        response=authed_session.get(cmd)
        result=json.loads(response.content)
        self.ctx.log("Operation name: {}".format(result["name"]))

        if (("done" in result.keys()) and result["done"]):
            self.ctx.log("Model training complete.")
            self.model_name = result["response"]["name"]
            self.ctx.log("Model full name: {}".format(self.model_name))
            self.ctx.config.set('model_name', self.model_name)
            self.ctx.config.write()
            response = self.client.list_model_evaluations(self.model_name)
            self.ctx.log("List of model evaluations:")
            for evaluation in response:
                self.ctx.log("Model evaluation name: {}".format(evaluation.name))
                self.ctx.log("Model evaluation id: {}".format(evaluation.name.split("/")[-1]))
                self.ctx.log("Model evaluation example count: {}".format(
                    evaluation.evaluated_example_count))
                self.ctx.log("Model evaluation time: {} seconds".format(evaluation.create_time.seconds))
                self.ctx.log("Full model evaluation: {}".format(inspect.getmembers(evaluation) ))
                self.ctx.log("\n")
        else:
            self.ctx.log("Model still training...")

    def deploy(self, model_id, locally=False):
        self.ctx.log('Google Deploy model ID: {}'.format(model_id))
        if (model_id is None):
            model_name = self.model_name
        else:
            model_name = self.client.model_path(self.project_id, self.compute_region, model_id)
        try:
            self.ctx.log('Deploy model: {}'.format(self.model_name))
            response = self.client.deploy_model(self.model_name)
            self.ctx.log("Deploy result: {}".format(response))
        except google.api_core.exceptions.FailedPrecondition as inst:
            self.ctx.log("Failed to deploy because its already deploying: {}...".format(inst))

    def predict(self, filename, model_id, threshold=None, locally=False):
        self.ctx.log('Google Predict')
        prediction_client = PredictionServiceClient()
        basefile, file_extension = os.path.splitext(filename)
        predictions_file = basefile+'_predicted.csv'
        self.ctx.log('Saving to file {}'.format(predictions_file))
        predictions=open(predictions_file, "wt")
        with open(filename,"rt") as csv_file:
            content = csv.reader(csv_file)
            next(content,None)
            csvlist = ''
            i=0
            for row in content:
                # Create payload
                values = []
                for column in row:
                    #self.ctx.log("Column: {}".format(column))
                    values.append({'number_value': float(column)})
                csvlist=",".join(row)
                print ("CSVList: {}".format(csvlist))
                payload = {
                    'row': {'values': values}
                }
                response = prediction_client.predict(self.model_name, payload)
                self.ctx.log("Prediction results:")
                for result in response.payload:
                    if ((threshold is None) or (result.classification.score >= score_threshold)):
                        prediction=result.tables.value.number_value
                self.ctx.log("Prediction: {}".format(prediction))
                csvlist += (',' + str(prediction) + '\n')
                predictions.write(csvlist)
                i = i + 1
        self.ctx.log('{} predictions.'.format(i))

    # TODO: rename to info and add to a2ml and for all providers    
    # def review(self):
    #     self.ctx.log('Google Review')

    #     # Get complete detail of the model.
    #     model = self.client.get_model(self.model_name)

    #     # Retrieve deployment state.
    #     if model.deployment_state == enums.Model.DeploymentState.DEPLOYED:
    #         deployment_state = "deployed"
    #     else:
    #         deployment_state = "undeployed"

    #     # Display the model information.
    #     self.ctx.log("Model name: {}".format(self.model_name))
    #     self.ctx.log("Model id: {}".format(model.name.split("/")[-1]))
    #     self.ctx.log("Model display name: {}".format(model.display_name))
    #     self.ctx.log("Model metadata:")
    #     self.ctx.log(model.tables_model_metadata)
    #     self.ctx.log("Model create time (seconds): {}".format(model.create_time.seconds))
    #     self.ctx.log("Model deployment state: {}".format(deployment_state))



import numpy as np
import pandas
from sklearn.metrics import make_scorer, recall_score, average_precision_score, roc_auc_score
from sklearn.metrics import confusion_matrix
from sklearn.metrics import matthews_corrcoef as mcc
from sklearn.metrics.scorer import SCORERS


def kappa(y_true, y_pred, weights=None, allow_off_by_one=False):
    """ https://en.wikipedia.org/wiki/Cohen%27s_kappa """
    # assert (len(y_true) == len(y_probs))
    try:
        if type(y_true) == pandas.DataFrame:
            y_true = [int(np.round(float(y))) for y in y_true.iloc[:,0].values]
        else:
            y_true = [int(np.round(float(y))) for y in y_true]

        if type(y_pred) == pandas.DataFrame:
            y_pred = [int(np.round(float(y))) for y in y_pred.iloc[:,0].values]
        else:
            y_pred = [int(np.round(float(y))) for y in y_pred]

    except ValueError as e:
        #   logger.error("For kappa, the labels should be integers or strings "
        #      "that can be converted to ints (E.g., '4.0' or '3').")
        raise

    # Figure out normalized expected values
    min_rating = min(min(y_true), min(y_pred))
    max_rating = max(max(y_true), max(y_pred))

    # shift the values so that the lowest value is 0
    # (to support scales that include negative values)
    y_true = [y - min_rating for y in y_true]
    y_pred = [y - min_rating for y in y_pred]

    # Build the observed/confusion matrix
    num_ratings = max_rating - min_rating + 1
    observed = confusion_matrix(y_true, y_pred, labels=list(range(num_ratings)))
    num_scored_items = float(len(y_true))

    # Build weight array if weren't passed one
    # if isinstance(weights, string_types):
    wt_scheme = weights
    weights = None
    # else:
    #    wt_scheme = ''
    if weights is None:
        weights = np.empty((num_ratings, num_ratings))
        for i in range(num_ratings):
            for j in range(num_ratings):
                diff = abs(i - j)
                if allow_off_by_one and diff:
                    diff -= 1
                if wt_scheme == 'linear':
                    weights[i, j] = diff
                elif wt_scheme == 'quadratic':
                    weights[i, j] = diff ** 2
                elif not wt_scheme:  # unweighted
                    weights[i, j] = bool(diff)
                else:
                    raise ValueError('Invalid weight scheme specified for '
                                     'kappa: {}'.format(wt_scheme))

    hist_true = np.bincount(y_true, minlength=num_ratings)
    hist_true = hist_true[: num_ratings] / num_scored_items
    hist_pred = np.bincount(y_pred, minlength=num_ratings)
    hist_pred = hist_pred[: num_ratings] / num_scored_items
    expected = np.outer(hist_true, hist_pred)

    # Normalize observed array
    observed = observed / num_scored_items

    # If all weights are zero, that means no disagreements matter.
    k = 1.0
    if np.count_nonzero(weights):
        k -= (sum(sum(weights * observed)) / sum(sum(weights * expected)))

    return k

def matthews(y_true, y_pred, sample_weight=None):
    return mcc(y_true, y_pred, sample_weight)

def gini(y_true, y_pred_proba, sample_weight=None):
    from sklearn.metrics import roc_auc_score
    roc_auc = roc_auc_score(y_true, y_pred_proba, sample_weight=sample_weight)
    return 2 * roc_auc - 1

def norm_macro_recall_score(y_true, y_pred):
    R = 0.5 #TODO: support multiclass R=(1/C) for C-class classification problems
    return (recall_score(y_true, y_pred, average='macro') - R)/(1 - R)

def average_precision_score_weighted_score(y_true, y_pred):
    return average_precision_score(y_true, y_pred, average='weighted')

def AUC_weighted_score(y_true, y_pred):
    return roc_auc_score(y_true, y_pred, average='weighted')

cohen_kappa_score = make_scorer(kappa)
matthews_corrcoef = make_scorer(matthews)
gini_score = make_scorer(gini, needs_threshold=True)
norm_macro_recall_scorer = make_scorer(norm_macro_recall_score)
average_precision_score_weighted_scorer = make_scorer(average_precision_score_weighted_score)
AUC_weighted_scorer = make_scorer(AUC_weighted_score)

SCORERS['cohen_kappa_score'] = cohen_kappa_score
SCORERS['matthews_corrcoef'] = matthews_corrcoef
SCORERS['gini'] = gini_score
SCORERS['norm_macro_recall'] = norm_macro_recall_scorer
SCORERS['average_precision_score_weighted'] = average_precision_score_weighted_scorer
SCORERS['AUC_weighted'] = AUC_weighted_scorer



import click
from a2ml.api.a2ml_experiment import A2MLExperiment
from a2ml.api.utils.context import pass_context


@click.group('experiment', short_help='Experiment management')
@pass_context
def cmdl(ctx):
    """Experiment management"""
    ctx.setup_logger(format='')


@click.command(short_help='List Experiments for selected DataSet')
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@pass_context
def list_cmd(ctx, provider):
    """List Experiments for selected DataSet"""
    A2MLExperiment(ctx, provider).list()


@click.command(short_help='Start Experiment')
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@pass_context
def start(ctx, provider):
    """Start Experiment.
       If Experiment is not selected, new Experiment will be created.
    """
    A2MLExperiment(ctx, provider).start()


@click.command(short_help='Stop Experiment')
@click.argument('run-id', required=False, type=click.STRING)
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@pass_context
def stop(ctx, provider, run_id):
    """Stop Experiment"""
    A2MLExperiment(ctx, provider).stop(run_id)


@click.command(short_help='Show Experiment leaderboard')
@click.argument('run-id', required=False, type=click.STRING)
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@pass_context
def leaderboard(ctx, provider, run_id):
    """Show Experiment leaderboard for specified run id.
       If run id is not provided, latest experiment run
       leaderboard will be shown.
    """
    A2MLExperiment(ctx, provider).leaderboard(run_id)

@click.command(short_help='Show Experiment history')
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@pass_context
def history(ctx, provider):
    """Show Experiment history"""
    A2MLExperiment(ctx, provider).history()


@pass_context
def add_commands(ctx):
    cmdl.add_command(list_cmd, name='list')
    cmdl.add_command(start)
    cmdl.add_command(stop)
    cmdl.add_command(leaderboard)
    cmdl.add_command(history)


add_commands()



import numpy as np
from sklearn.metrics import make_scorer, mean_squared_error, mean_squared_log_error, mean_absolute_error
from sklearn.metrics.scorer import SCORERS


def neg_rmsle_score(y_true, y_pred):
    return -np.sqrt(mean_squared_log_error(y_true, y_pred))


def neg_rmse_score(y_true, y_pred):
    return -np.sqrt(mean_squared_error(y_true, y_pred))


def neg_mase_score(y_true, y_pred):
    y_true = np.asarray(y_true)
    y_pred = np.asarray(y_pred)
    numerator = np.sum(np.abs(y_true - y_pred))
    coeff = y_true.shape[0] / (y_true.shape[0] - 1)
    denominator = np.sum(np.abs(y_true[1:] - y_true[:-1]))
    if np.abs(denominator) < 1e-12:     # straight line
        denominator = 1.0
    return -numerator / (denominator * coeff)


def neg_mape_score(y_true, y_pred):
    eps = 1e-8
    y_true = np.asarray(y_true)
    y_pred = np.asarray(y_pred)
    result = -np.sum(np.abs((y_true - y_pred) / (y_true + eps) )) / len(y_true)
    return float(result)


def mda_score(y_true, y_pred, above_percent=0.2):
    """ https://en.wikipedia.org/wiki/Mean_Directional_Accuracy """
    y_true = np.asarray(y_true)
    y_pred = np.asarray(y_pred)

    y_diff = y_true[1:] - y_true[:-1]
    y_diff_mean = np.mean(y_diff)

    idx = np.where(y_diff / y_diff_mean > above_percent)[0]
    true_sign = np.sign(y_diff[idx])

    y_diff = y_pred[1:] - y_true[:-1]
    idx_eliminate = np.where(y_diff / y_diff_mean > above_percent)[0]
    y_diff[idx_eliminate] = 0.
    pred_sign = np.sign(y_diff[idx])

    match = np.sum(true_sign == pred_sign)
    return match / float(idx.shape[0])

def nmae_score(y_true, y_pred):
    """ Normalized Root Mean Squared Error """
    return mean_absolute_error(y_true, y_pred) / (y_true.max() - y_true.min())

def nrmse_score(y_true, y_pred):
    """ Normalized Root Mean Squared Error """
    return np.sqrt(mean_squared_error(y_true, y_pred)) / (y_true.max() - y_true.min())

def spearman_correlation_score(y_true, y_pred):
    from scipy import stats

    return stats.spearmanr(y_true, y_pred)[0]

neg_rmsle_scorer = make_scorer(neg_rmsle_score)
neg_mase_scorer = make_scorer(neg_mase_score)
neg_mape_scorer = make_scorer(neg_mape_score)
mda_scorer = make_scorer(mda_score)
neg_rmse_scorer = make_scorer(neg_rmse_score)
nmae_scorer = make_scorer(nmae_score)
nrmse_scorer = make_scorer(nrmse_score)
spearman_correlation_scorer = make_scorer(spearman_correlation_score)

SCORERS['neg_rmsle'] = neg_rmsle_scorer
SCORERS['neg_mase'] = neg_mase_scorer
SCORERS['neg_mape'] = neg_mape_scorer
SCORERS['mda'] = mda_scorer
SCORERS['neg_rmse'] = neg_rmse_scorer
SCORERS['normalized_mean_absolute_error'] = nmae_scorer
SCORERS['normalized_root_mean_squared_error'] = nrmse_scorer
SCORERS['spearman_correlation'] = spearman_correlation_scorer



from .base import AugerBaseApi
from ..exceptions import AugerException


class AugerOrganizationApi(AugerBaseApi):
    """Auger Organization API."""

    def __init__(self, ctx, org_name=None, org_id=None):
        super(AugerOrganizationApi, self).__init__(
            ctx, None, org_name, org_id)

    # def get_cluster_mode(self):
    #     cluster_mode = getattr(self, 'cluster_mode', None)
    #     if not cluster_mode:
    #         self.cluster_mode = self.properties().get('cluster_mode')
    #     return self.cluster_mode

    def create(self):
        raise AugerException(
            'You could\'t create organization using Auger Cloud API.'
            ' Please use Auger UI to do that...')

    def delete(self):
        raise AugerException(
            'You could\'t delete organization using Auger Cloud API.'
            ' Please use Auger UI to do that...')



import os
import subprocess
import numpy as np

from ..cloud.cluster import AugerClusterApi
from ..cloud.pipeline import AugerPipelineApi
from ..cloud.endpoint import AugerEndpointApi
from ..cloud.review_alert import AugerReviewAlertApi
from ..cloud.review_alert_item import AugerReviewAlertItemApi
from ..exceptions import AugerException
from ..cloud.pipeline_file import AugerPipelineFileApi
from a2ml.api.utils import fsclient


class ModelDeploy(object):
    """Deploy Model on locally or on Auger Cloud."""

    def __init__(self, ctx, project):
        super(ModelDeploy, self).__init__()
        self.project = project
        self.ctx = ctx

    def execute(self, model_id, locally=False, review=True):
        if locally:
            return self.deploy_model_locally(model_id, review)
        else:
            return self.deploy_model_in_cloud(model_id, review)

    def create_update_review_alert(self, model_id, pipeline_properties=None, parameters=None):
        if not pipeline_properties:
            pipeline_properties = AugerPipelineApi(self.ctx, None, model_id).properties()

        endpoint_api = None    
        if not pipeline_properties.get('endpoint_pipelines'):
            self.ctx.log('Creating review endpoint ...')
            endpoint_api = AugerEndpointApi(self.ctx, None)
            endpoint_properties = endpoint_api.create(pipeline_properties.get('id'))
            pipeline_properties['endpoint_pipelines']= [endpoint_properties.get('id')]
        else:
            endpoint_api = AugerEndpointApi(self.ctx, None, 
                pipeline_properties['endpoint_pipelines'][0].get('endpoint_id'))

        AugerReviewAlertApi(self.ctx, endpoint_api).create_update(parameters)

    def review(self, model_id):
        pipeline_properties = AugerPipelineApi(self.ctx, None, model_id).properties()

        result = {}
        if not pipeline_properties.get('endpoint_pipelines'):
            return result

        endpoint_api = AugerEndpointApi(self.ctx, None, 
            pipeline_properties['endpoint_pipelines'][0].get('endpoint_id'))

        alert_items = AugerReviewAlertItemApi(self.ctx, endpoint_api).list()
        if not alert_items:
            return result

        alert_items = sorted(alert_items, key=lambda k: k['id'], reverse=True)
        alert_item = alert_items[0]
        error_states = ['review_data_build_failed', 'project_file_processing_failed', 'experiment_session_failed', 'pipeline_creating_failed']

        alert = AugerReviewAlertApi(self.ctx, None, alert_item['review_alert_id']).properties()
        status = "started"
        error = ""

        if alert.get('actions') == 'retrain_deploy':
            retrain_status = alert_item.get('action_results', {}).get('retrain')
            if retrain_status in error_states:
                status = 'error'
                error = retrain_status
            else:            
                redeploy_status = alert_item.get('action_results', {}).get('redeploy')
                if redeploy_status in error_states:
                    status = 'error'
                    error = redeploy_status
                elif redeploy_status == 'endpoint_updated' or redeploy_status == 'endpoint_has_better_pipeline':
                    status = 'completed'
        elif alert.get('actions') == 'retrain':
            retrain_status = alert_item.get('action_results', {}).get('retrain')
            if retrain_status in error_states:
                status = 'error'
                error = retrain_status
            elif retrain_status == 'experiment_session_done':
                status = 'completed'
        else:
            status = 'completed'

        accuracy = 0
        if alert_item.get('values'):
            accuracy = np.mean(alert_item['values'])
        result = {
            'status': status,
            'error': error,
            'accuracy': accuracy
        }
        return result
            
    def deploy_model_in_cloud(self, model_id, review):
        self.ctx.log('Deploying model %s' % model_id)

        self.project.start()

        pipeline_properties = AugerPipelineApi(
            self.ctx, None).create(model_id, review)

        if pipeline_properties.get('status') == 'ready':
            if review:
                self.create_update_review_alert(model_id, pipeline_properties)

            self.ctx.log('Deployed Model on Auger Cloud. Model id is %s' % \
                pipeline_properties.get('id'))            
        else:
            self.ctx.log('Deployed Model on Auger Cloud failed. Model id is %s.Error: %s' % \
                (pipeline_properties.get('id'), pipeline_properties.get('error_message', "")))

        return pipeline_properties.get('id')

    def deploy_model_locally(self, model_id, review):
        is_loaded, model_path, model_name = self.verify_local_model(model_id)
        #TODO: support review flag
        if not is_loaded:
            self.ctx.log('Downloading model %s' % model_id)

            self.project.start()

            pipeline_file_api = AugerPipelineFileApi(self.ctx, None)
            pipeline_file_properties = pipeline_file_api.create(model_id)
            downloaded_model_file = pipeline_file_api.download(
                pipeline_file_properties['signed_s3_model_path'],
                model_path, model_id)

            self.ctx.log('Downloaded model to %s' % downloaded_model_file)

            self.ctx.log('Pulling docker image required to predict')
            self._docker_pull_image()            
        else:
            self.ctx.log('Downloaded model is %s' % model_name)

        return model_id

    def verify_local_model(self, model_id):
        model_path = os.path.join(self.ctx.config.get_path(), 'models')
        model_name = os.path.join(model_path, 'model-%s.zip' % model_id)
        is_exists = fsclient.is_folder_exists(os.path.join(model_path,"model-%s"%model_id))
        if not is_exists:
            is_exists = fsclient.is_file_exists(model_name)
        return is_exists, model_path, model_name

    def _docker_pull_image(self):
        cluster_settings = AugerClusterApi.get_cluster_settings(self.ctx)
        docker_tag = cluster_settings.get('kubernetes_stack')

        try:
            subprocess.check_call(
                'docker pull deeplearninc/auger-ml-worker:%s' % \
                 docker_tag, shell=True)
        except subprocess.CalledProcessError as e:
            raise AugerException('Can\'t pull Docker container...')

        return docker_tag



from a2ml.api.base_a2ml import BaseA2ML
from a2ml.api.utils.show_result import show_result
from a2ml.api.utils import convert_source


class A2ML(BaseA2ML):
    """Facade to A2ML providers."""

    def __init__(self, ctx, provider = None):
        """Initializes A2ML PREDIT instance.

        Args:
            ctx (object): An instance of the a2ml Context.
            provider (str): The automl provider(s) you wish to run. For example 'auger,azure,google'. The default is None - use provider set in config.

        Returns:
            A2ML object

        Examples:
            .. code-block:: python

                ctx = Context()
                a2ml = A2ML(ctx, 'auger, azure')
        """
        super(A2ML, self).__init__(ctx, None)
        self.runner = self.build_runner(ctx, provider)
        self.local_runner = lambda: self.build_runner(ctx, provider, force_local=True)

    @show_result
    def import_data(self, source=None):
        """Imports data defined in context. Uploading the same file name will result in versions being appended to the file name.

        Note:
            Your context points to a config file where ``source`` is defined.

            .. code-block:: yaml

                # Local file name or remote url to the data source file
                source: './dataset.csv'

        Args:
            source (str, optional): Local file name or remote url to the data source file or Pandas DataFrame

        Returns:
            Results for each provider. ::

                {
                    'auger': {'result': True, 'data': {'created': 'dataset.csv'}}, \n
                    'azure': {'result': True, 'data': {'created': 'dataset.csv'}}
                }

            Errors. ::

                {
                    'auger': {'result': False, 'data': 'Please specify data source file...'}, \n
                    'azure': {'result': False, 'data': 'Please specify data source file...'}
                }


        Examples:
            .. code-block:: python

                ctx = Context()
                a2ml = A2ML(ctx, 'auger, azure')
                a2ml.import_data()
        """
        with convert_source(source, self.ctx.config.get("name", "source_data")) as data_source:
            return self.runner.execute('import_data', source=data_source)

    @show_result
    def train(self):
        """Starts training session based on context state.

        Returns:
            Results for each provider. ::

                {
                    'auger': {
                        'result': True,
                        'data': {
                            'experiment_name': 'dataset.csv-4-experiment',
                            'session_id': '9ccfe04eca67757a'
                         }
                    },
                    'azure': {
                        'result': True,
                        'data': {
                            'experiment_name': 'dataset.csv-4-experiment',
                            'session_id': '9ccfe04eca67757a'
                         }
                    }
                }

            Errors. ::

                {
                    'auger': {'result': False, 'data': 'Please set target to build model.'}, \n
                    'azure': {'result': False, 'data': 'Please set target to build model.'}
                }

        Examples:

            .. code-block:: python

                ctx = Context()
                a2ml = A2ML(ctx, 'auger, azure')
                a2ml.train()

        """
        return self.runner.execute('train')

    @show_result
    def evaluate(self, run_id = None):
        """Evaluate the results of training.

        Args:
            run_id (str, optional): The run id for a training session. A unique run id is created for every train. Default is last experiment train.
        
        Returns:
            Results for each provider. ::

                {
                    'auger': {
                        'result': True,
                        'data': {
                            'run_id': '9ccfe04eca67757a',
                            'leaderboard': [
                                {'model id': 'A017AC8EAD094FD', 'rmse': '0.0000', 'algorithm': 'LGBMRegressor'},
                                {'model id': '4602AFCEEEAE413', 'rmse': '0.0000', 'algorithm': 'ExtraTreesRegressor'}
                            ],
                            'trials_count': 10,
                            'status': 'started',
                            'provider_status': 'provider specific'
                        }
                    },
                    'azure': {
                        'result': True,
                        'data': {
                            'run_id': '9ccfe04eca67757a',
                            'leaderboard': [
                                {'model id': 'A017AC8EAD094FD', 'rmse': '0.0000', 'algorithm': 'LGBMRegressor'},
                                {'model id': '4602AFCEEEAE413', 'rmse': '0.0000', 'algorithm': 'ExtraTreesRegressor'}
                            ],
                            'trials_count': 10,
                            'status': 'started',
                            'provider_status': 'provider specific'                            
                        }
                    }                    
                }

            **Status**

                * **preprocess** - search is preprocessing data for traing
                * **started** - search is in progress
                * **completed** - search is completed
                * **interrupted** - search was interrupted
                * **error** - search was finished with error

        Examples:
            .. code-block:: python

                ctx = Context()
                a2ml = A2ML(ctx, 'auger, azure')
                while True:
                    res = a2ml.evaluate()
                    if status['auger']['status'] not in ['preprocess','started']:
                        break
        """
        return self.runner.execute('evaluate', run_id = run_id)

    @show_result
    def deploy(self, model_id, locally=False, review=True, provider=None):
        """Deploy a model locally or to specified provider(s).

        Note:
            See evaluate function to get model_id \n
            This method support only one provider

        Args:
            model_id (str): The model id from any experiment you will deploy.
            locally(bool): Deploys the model locally if True, on the Provider Cloud if False. The default is False.
            review(bool): Should model support review based on actual data. The default is True.
            provider (str): The automl provider you wish to run. For example 'auger'. The default is None - use provider defined by model_id or set in costructor.

        Returns:
            ::

                {
                    'result': True,
                    'data': {'model_id': 'A017AC8EAD094FD'}
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                a2ml = A2ML(ctx, 'auger, azure')
                a2ml.deploy(model_id='A017AC8EAD094FD')
        """
        return self.get_runner(locally, model_id, provider).execute_one_provider('deploy', model_id, locally, review)

    @show_result
    def predict(self, filename,
      model_id, threshold=None, locally=False, data=None, columns=None, predicted_at=None, output=None, provider=None):
        """Predict results with new data against deployed model. Predictions are stored next to the file with data to be predicted on. The file name will be appended with suffix _predicted.

        Note:
            Use deployed model_id \n
            This method support only one provider
            If you pass prediction_id column in input data, it will be return in predict result. Otherwise unique prediction_id will be generated for each record.

        Args:
            filename(str): The file with data to request predictions for.
            model_id(str): The deployed model id you want to use.
            threshold(float): For classification models only. This will return class probabilities with response.
            locally(bool): Predicts using a local model if True, on the Provider Cloud if False.
            data: dict or array of records or Pandas DataFrame
            columns(list): list of column names if data is array of records
            predicted_at: Predict data date. Use for review of historical data.
            output(str): Output csv file path.
            provider (str): The automl provider you wish to run. For example 'auger'. The default is None - use provider set in costructor or config.

        Returns:
            if filename is not None. ::

                {
                    'result': True,
                    'data': {'predicted': 'dataset_predicted.csv'}
                }

            if filename is None and data is not None and columns is None. ::

                {
                    'result': True,
                    'data': {'predicted': [{'prediction_id': 123, col1: value1, col2: value2}, {'prediction_id': 456, col1: value3, col2: value4}]}
                }

            if filename is None and data is not None and columns is not None. ::

                {
                    'result': True,
                    'data': {'predicted': {'columns': ['prediction_id', 'col1', 'col2'], 'data': [['789', 'value1', 'value2'], ['321', 'value3', 'value4']]}}
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                rv = A2ML(ctx).predict('../irises.csv', model_id)
                # if rv[provider].result is True
                # predictions are stored in rv[provider]['data']['predicted']

            .. code-block:: python

                ctx = Context()
                data = [{'col1': 'value1', 'col2': 'value2'}, {'col1': 'value3', 'col2': 'value4'}]
                rv = A2ML(ctx).predict(None, model_id, data=data)
                # if rv[provider].result is True
                # predictions are returned as rv[provider]['data']['predicted']

            .. code-block:: python

                ctx = Context()
                data = [['value1', 'value2'], ['value3', 'value4']]
                columns = ['col1', 'col2']                
                rv = A2ML(ctx).predict(None, model_id, data=data)
                # if rv[provider].result is True
                # predictions are returned as rv[provider]['data']['predicted']

        """
        return self.get_runner(locally, model_id, provider).execute_one_provider('predict', filename, model_id, threshold, locally, data, columns, predicted_at, output)

    @show_result
    def actuals(self, model_id, filename=None, actual_records=None, actuals_at=None, locally=False, provider=None):
        """Submits actual results(ground truths) for predictions of a deployed model. This is used to review and monitor active models.

        Note:
            It is assumed you have predictions against this model first. The file will need to fill in actual values for prediction_id.

            .. list-table:: actuals.csv
                :widths: 50 50
                :header-rows: 1

                * - prediction_id
                  - actual
                * - eaed9cd8-ba49-4c06-86d5-71d453c681d1
                  - Iris-setosa
                * - eaed9cd8-ba49-4c06-86d5-71d453c65290
                  - Iris-virginica

            This method support only one provider

        Args:
            model_id(str): The deployed model id you want to use.
            filename(str): The file with data to request predictions for.
            actual_records: array of records [[prediction_id, actual]] or Pandas DataFrame (prediction_id, actual)
            actuals_at: Actuals date. Use for review of historical data.
            locally(bool): Process actuals locally.
            provider (str): The automl provider you wish to run. For example 'auger'. The default is None - use provider set in costructor or config.

        Returns:
            ::

                {
                    'result': True,
                    'data': True
                }

            Errors. ::

                {
                    'result': False,
                    'data': 'Actual Prediction IDs not found in model predictions.'
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                model = A2MLModel(ctx).actuals('D881079E1ED14FB', filename=<path_to_file>/actuals.csv)

            .. code-block:: python
            
                # To pass just one actual:
                ctx = Context()
                actual_records = [['prediction_id', 'actual_value']]
                model = A2ML(ctx).actuals('D881079E1ED14FB', actual_records=actual_records)

            .. code-block:: python

                ctx = Context()
                actual_records = [['prediction_1', 'value1'], ['prediction_2', 'value2']]
                model = A2ML(ctx).actuals('D881079E1ED14FB', actual_records=actual_records)

        """
        return self.get_runner(locally, model_id, provider).execute_one_provider('actuals', model_id, filename, actual_records, actuals_at, locally)


    @show_result
    def delete_actuals(self, model_id, with_predictions=False, begin_date=None, end_date=None, locally=False, provider=None):
        """Delete files with actuals and predcitions locally or from specified provider(s).

        Args:
            model_id (str): Model ID to delete actuals and predictions.
            with_predictions(bool): 
            begin_date: Date to begin delete operations 
            end_date: Date to end delete operations 
            locally(bool): Delete files from local model if True, on the Provider Cloud if False. The default is False.
            provider (str): The automl provider you wish to run. For example 'auger'. The default is None - use provider defined by model_id or set in costructor.

        Returns:
            ::

                {
                    'result': True,
                    'data': None
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                A2MLModel(ctx).delete_actuals(model_id='D881079E1ED14FB')
        """
        return self.get_runner(locally, model_id, provider).execute_one_provider('delete_actuals', model_id, with_predictions, begin_date, end_date, locally)


    @show_result
    def review(self, model_id, locally=False, provider=None):
        """Review information about deployed model.

        Args:
            model_id(str): The deployed model id you want to use.
            locally(bool): Process review locally.

        Returns:
            ::

                {
                    'result': True,
                    'data': {'status': 'completed', 'error': ''}
                }

        Examples:
            .. code-block:: python

                ctx = Context()
                model = A2ML(ctx).review(model_id='D881079E1ED14FB')
        """
        return self.get_runner(locally, model_id, provider).execute_one_provider('review', model_id)



import uuid
import dateutil
from collections import OrderedDict
from collections.abc import Iterable
import time
import os
import logging
import pandas as pd
import contextlib

from a2ml.api.utils import fsclient


def to_list(obj):
    if obj:
        if isinstance(obj, Iterable) and not isinstance(obj, str):
            return list(obj)
        else:
            return [obj]
    else:
        return []

def dict_dig(obj, *path):
    curr = obj
    i = 0

    while curr and i < len(path) and isinstance(curr, dict):
        curr = curr.get(path[i], None)
        i += 1

    return curr

def get_uid():
    return uuid.uuid4().hex[:15].upper()

def get_uid4():
    return str(uuid.uuid4())

def remove_dups_from_list(ar):
    return list(OrderedDict.fromkeys(ar))

def process_arff_line(line, date_attrs):
    if "@attribute" in line.lower():
        parts = line.split(maxsplit=3)
        if len(parts) > 2 and parts[2].lower() == 'date':
            line = parts[0]+ ' ' + parts[1] + ' string\n'
            date_field = parts[1].strip("\"\'")
            date_format = parts[3].strip("\"\'\n")

            date_attrs[date_field] = date_format

    return line

def url_encode(path):
    try:
        from urllib.parse import urlparse, parse_qs, quote
    except ImportError:
        from urlparse import urlparse, parse_qs, quote

    return quote(path, safe='#&%:/?*=\'')

def get_remote_file_info(remote_path):
    from urllib.request import urlopen
    import urllib
    import codecs

    try:
        from urllib.parse import urlparse, parse_qs
    except ImportError:
        from urlparse import urlparse, parse_qs

    file_info  = {}
    remote_path1 = url_encode(remote_path)
    # logging.info("Get info for file from url: %s" % remote_path1)

    req = urlopen(remote_path1)
    #print(req.info())

    # Make local path
    contentDisp = req.getheader('Content-Disposition')
    contentType = req.getheader('Content-Type')
    fileext = ''
    if contentType:
        if contentType == 'application/x-gzip':
            fileext = '.gz'
        elif contentType == 'text/csv':
            fileext = '.csv'

    filename = ''
    if contentDisp:
        items = contentDisp.split(';')
        for item in items:
            item = item.strip()
            if item.startswith("filename=\""):
                filename = item[10:-1]
                break
            elif item.startswith("filename="):
                filename = item[9:]
                break

    if not filename:
        uri = urlparse(remote_path)
        params = parse_qs(uri.query)
        if len(params.get('id', []))>0:
            filename = params['id'][0]+fileext
        else:
            if len(uri.path)>0 and len(os.path.basename(uri.path))>0:
                filename = os.path.basename(uri.path)
            else:
                filename = get_uid()+fileext

    remote_file_size = 0
    try:
        remote_file_size = int(req.getheader('Content-Length'))
    except:
        pass

    remote_file_etag = ''
    try:
        remote_file_etag = req.getheader('ETag').strip("\"\'")
    except:
        pass

    req.close()

    file_info = {
        'file_name': "",
        'file_ext': "",
        'file_size': remote_file_size,
        'file_etag': remote_file_etag
    }

    dot_index = filename.find('.')
    if dot_index>=0:
        file_info['file_name'] = filename[:dot_index]
        file_info['file_ext'] = filename[dot_index:]
    else:
        file_info['file_name'] = filename

    return file_info

def download_file(remote_path, local_dir, file_name, force_download=False):
    local_file_path = ""
    download_file = True
    remote_file_info = {}

    #logging.info("download_file: %s, %s, %s, %s"%(remote_path, local_dir, file_name, force_download))
    if file_name:
        all_local_files = fsclient.list_folder(os.path.join(local_dir, file_name+".*"), wild=True, remove_folder_name=True)
        if all_local_files:
            local_file_path = os.path.join( local_dir, all_local_files[0])

    if not local_file_path:
        remote_file_info = get_remote_file_info(remote_path)
        if not remote_file_info:
            raise Exception("Remote path does not exist or unaccessible: %s"%(remote_path))

        if  file_name:
            local_file_path = os.path.join(local_dir,
                file_name+remote_file_info.get('file_ext'))
        else:
            local_file_path = os.path.join(local_dir,
                remote_file_info.get('file_name') + remote_file_info.get('file_ext'))

    if fsclient.is_file_exists(local_file_path):
        etag_changed = False
        file_size_changed = False

        if force_download:
            logging.info("Force download file again.")

        if force_download or etag_changed or file_size_changed:
            fsclient.remove_file(local_file_path)
        else:
            download_file = False

    if download_file:
        #logging.info("Download to local file path: %s"%local_file_path)
        fsclient.download_file(remote_path, local_file_path)

    return local_file_path

# by default value from other dict overwrites value in d
def merge_dicts(d, other, concat_func=lambda v, ov: ov):
    from collections.abc import Mapping

    for k, v in other.items():
        d_v = d.get(k)
        if isinstance(v, Mapping) and isinstance(d_v, Mapping):
            merge_dicts(d_v, v, concat_func)
        else:
            if k in d:
                d[k] = concat_func(d[k], v)
            else:
                d[k] = v

    return d

def convert_to_date(date):
    if type(date) is str:
        return dateutil.parser.parse(date).date()
    elif type(date) is time.time:
        return date.date()
    else:
        return date

@contextlib.contextmanager
def convert_source(source, name):
    if source is not None and isinstance(source, pd.DataFrame):
        with fsclient.save_atomic("%s.parquet"%name, move_file=False) as local_path:
            source.to_parquet(local_path, index=False, compression="gzip")
            yield local_path
    else:
        yield source

def retry_helper(func, retry_errors=[], num_try=10, delay=10, ctx=None):
    nTry = 0
    while True:
        try:
            return func()
        except Exception as exc:
            retry_exc = False
            if not retry_errors:
                retry_exc = True
            else:    
                for retry_error in retry_errors:
                    if retry_error in str(exc):
                        retry_exc = True
                        break

            if retry_exc and nTry < num_try:
                if ctx:
                    ctx.log("Retry '%s', error: '%s'. Sleep and try again. Num try: %s"%(func, str(exc), nTry))
                else:
                    logging.error("Retry '%s', error: '%s'. Sleep and try again. Num try: %s"%(func, str(exc), nTry))

                nTry += 1
                time.sleep(delay*nTry)
            else:
                raise                
                



from .base import AugerBaseApi
from ..exceptions import AugerException


class AugerPredictionApi(AugerBaseApi):
    """Auger Trial API."""

    def __init__(self, ctx, pipeline_api,
        prediction_name=None, prediction_id=None, use_endpoint=False):
        super(AugerPredictionApi, self).__init__(
            ctx, pipeline_api, prediction_name, prediction_id)
        assert pipeline_api is not None, 'Pipeline must be set for Prediction'

        self.use_endpoint = use_endpoint
        if self.use_endpoint:
            self.parent_id_name = "endpoint_id"
            self._set_api_request_path("AugerEndpointPredictionApi")

    def create(self, records, features, threshold=None, file_url=None, predicted_at=None):
        params = {
            'records': records, 
            'features': features,
            'threshold': threshold
        }
        if self.use_endpoint:
            params['endpoint_id'] = self.parent_api.object_id
            if not self.ctx.config.get('support_model_history'):
                params['primary_only'] = True
        else:
            params['pipeline_id'] = self.parent_api.object_id

        if file_url:
            params['file_url'] = file_url
                
        if predicted_at:
            params['predicted_at'] = str(predicted_at)

        return self._call_create(params, ['requested', 'running'])



import importlib

class CRUDRunner(object):
    """Runner executes provider crud operations."""

    def __init__(self, ctx, providers, obj_name):
        super(CRUDRunner, self).__init__()
        self.ctx = ctx
        self.providers = self._load_providers(providers, obj_name)

    def execute_one_provider(self, operation_name, *args, **kwargs):
        result = self.execute(operation_name, *args, **kwargs)        
        return list(result.values())[0]

    def execute(self, operation_name, *args, **kwargs):
        def send_error(p, msg):
            self.ctx.log('[%s]  %s' % (p, msg))
            raise Exception(msg)

        if len(self.providers) == 0:
            msg = 'Please specify provider(s) to run %s' % operation_name
            self.ctx.log(msg)
            return {'error': msg}

        results = {}
        for p in self.providers.keys():
            pi = self.providers[p]
            try:
                try:
                    if isinstance(pi, str):
                        send_error(p, 'Can\'t load %s to run \'%s\''\
                            % (pi, operation_name))
                    else:
                        result = getattr(pi, operation_name)(*args, **kwargs)
                        results[p] = {
                            'result': True,
                            'data': result }
                except AttributeError:
                    send_error(p, '%s not found' % operation_name)
            except Exception as e:
                if self.ctx.debug:
                    import traceback
                    traceback.print_exc()
                results[p] = {
                    'result': False,
                    'data': str(e) }

        return results

    def _load_providers(self, provider, obj_name):
        def get_provider_class(p):
            module = importlib.import_module('a2ml.api.%s.%s' % (p, obj_name))
            return getattr(module, '%s%s' % (p.capitalize(),obj_name.capitalize()))

        def get_instance(p):
            new_ctx = self.ctx.copy(p)
            
            if self.ctx.use_auger_cloud() and p != 'auger':
                new_ctx.provider_info = {p: {"project": {
                    "cluster": self.ctx.config.get("cluster", config_name=p),
                    "deploy_cluster": self.ctx.config.get("deploy_cluster", config_name=p)
                }}}
                provider_class = get_provider_class("auger")
            else:
                provider_class = get_provider_class(p)

            return provider_class(new_ctx)

        return {p: get_instance(p) for p in self.ctx.get_providers(provider)}




from .api.a2ml import A2ML
from .api.a2ml_project import A2MLProject
from .api.a2ml_dataset import A2MLDataset
from .api.a2ml_experiment import A2MLExperiment
from .api.a2ml_model import A2MLModel
from .api.utils.context import Context



from .base import AugerBaseApi
from .dataset import AugerDataSetApi
from a2ml.api.auger.dataset import AugerDataset
from .experiment_session import AugerExperimentSessionApi
from ..exceptions import AugerException

MODEL_TYPES = ['classification', 'regression', 'timeseries']

class AugerExperimentApi(AugerBaseApi):
    """Auger Experiment Api."""

    def __init__(self, ctx, project_api,
        experiment_name=None, experiment_id=None):
        super(AugerExperimentApi, self).__init__(
            ctx, project_api, experiment_name, experiment_id)
        assert project_api is not None, 'Project must be set for Experiment'
        self._set_api_request_path('AugerExperimentApi')

    def run(self):
        experiment_session_api = \
            AugerExperimentSessionApi(self.ctx, self)
        experiment_session_properties = \
            experiment_session_api.create()
        experiment_session_api.run()
        return experiment_session_properties.get('id')

    def create(self, data_set_name):
        assert data_set_name is not None, \
            'DataSet Name is required to create Experiment'

        data_set_api = AugerDataSetApi(
            self.ctx, self.parent_api, data_set_name)
        data_set_properties = data_set_api.properties()

        if not self.object_name:
            self.object_name = self._get_uniq_object_name(
                data_set_name, '-experiment')

        return self._call_create({
            'name': self.object_name,
            'project_id': self.parent_api.object_id,
            'data_path': data_set_properties.get('url')})

    def get_experiment_settings(self):
        config = self.ctx.config

        model_type = config.get('model_type', '')
        if not model_type in MODEL_TYPES:
            raise AugerException('Model type should be %s' % \
                '|'.join(MODEL_TYPES))

        test_data_path = None
        if config.get('experiment/validation_source'):
            data_set_api = None
            if self.ctx.config.get('experiment/validation_dataset'):
                data_set_api = AugerDataSetApi(self.ctx, self.parent_api, config.get('experiment/validation_dataset'))
                data_set_properties = data_set_api.properties()
                test_data_path = data_set_properties.get('url')

            if not test_data_path:
                data_set_api = AugerDataset(self.ctx)._create( self.parent_api,
                    source = self.ctx.config.get('experiment/validation_source'),
                    validation = True
                )
                data_set_properties = data_set_api.properties()
                test_data_path = data_set_properties.get('url')
        else:
            self.ctx.config.remove('experiment/validation_dataset')
            self.ctx.config.write()

        options = {
            'crossValidationFolds':
                config.get('experiment/cross_validation_folds', 5),
            'max_total_time_mins':
                config.get('experiment/max_total_time', 60),
            'max_eval_time_mins':
                config.get('experiment/max_eval_time', 6),
            'max_n_trials':
                config.get('experiment/max_n_trials', 1000),
            'use_ensemble':
                config.get('experiment/use_ensemble', True),
            'classification':
                True if model_type == 'classification' else False,
            'scoring':
                config.get('experiment/metric',
                    'accuracy' if model_type == 'classification' else 'r2'),
            'test_data_path': test_data_path
        }

        if config.get('experiment/trials_per_worker'):
            options['trials_per_worker'] = config.get('experiment/trials_per_worker')
        if config.get('experiment/class_weight'):
            options['algorithm_params_common'] = {'class_weight': config.get('experiment/class_weight')}
        if config.get('experiment/oversampling'):
            options['oversampling'] = config.get('experiment/oversampling')
        if config.get('experiment/estimate_trial_time'):
            options['apply_estimate_trial_time'] = config.get('experiment/estimate_trial_time')
        if config.get('experiment/max_cores_per_trial'):
            options['cpu_per_mt_algorithm'] = config.get('experiment/max_cores_per_trial')
        if config.get('experiment/max_concurrent_trials'):
            options['trials_per_worker'] = config.get('experiment/max_concurrent_trials')
        if self.ctx.config.get('experiment/blocked_models'):
            options["algorithms_to_exlude"] = self.ctx.config.get_list('experiment/blocked_models')
        if self.ctx.config.get('experiment/allowed_models'):
            options["allowed_algorithms"] = self.ctx.config.get_list('experiment/allowed_models')

        data_set_id = self.properties()['project_file_id']
        data_set_api = AugerDataSetApi(
            self.ctx, self.parent_api, None, data_set_id)
        data_set_properties = data_set_api.properties()
        stats = data_set_properties['statistics']

        self._fill_data_options(model_type, stats)

        return {'evaluation_options': options}, model_type, stats

    def _fill_data_options(self, model_type, stats):
        config = self.ctx.config
        target = config.get('target')
        if not target:
            raise AugerException('Please set target to build model.')

        exclude = config.get_list('exclude', [])
        label_encoded = config.get_list('experiment/label_encoded', [])
        categoricals = config.get_list('experiment/categoricals', [])
        date_time = config.get_list('experiment/date_time', [])
        time_series = None
        if model_type == 'timeseries':
            time_series = config.get('experiment/time_series', None)
            if not time_series:
                raise AugerException('Please select time series feature'
                    ' to build time series model'
                    ' (experiment/time_series option).')

        for item in stats.get('stat_data'):
            column_name = item['column_name']
            item['use'] = not column_name in exclude and column_name != target
            item['isTarget'] = target == column_name

            if column_name in label_encoded:
                item['datatype'] = 'hashing'
            elif column_name in categoricals:
                item['datatype'] = 'categorical'
            elif column_name in date_time:
                item['datatype'] = 'datetime'
            elif column_name == time_series:
                item['datatype'] = 'timeseries'
